import math
from dipy.utils.optpkg import optional_package
from dipy.utils.tripwire import TripWire

keras_msg = (
    "To use afqinsight's tensorflow models, you need to have tensorflow "
    "installed. You can do this by installing afqinsight with `pip install "
    "afqinsight[tf]`, or by separately installing tensorflow with `pip install "
    "tensorflow`."
)

tf, has_tf, _ = optional_package("tensorflow", trip_msg=keras_msg)

if has_tf:
    from tensorflow.keras.models import Model
    from tensorflow.keras.layers import Dense, Flatten, Dropout, Input
    from tensorflow.keras.layers import MaxPooling1D, Conv1D
    from tensorflow.keras.layers import LSTM, Bidirectional
    from tensorflow.keras.layers import (
        BatchNormalization,
        GlobalAveragePooling1D,
        Permute,
        concatenate,
        Activation,
        add,
    )
else:
    # Since all model building functions start with Input, we make Input the
    # tripwire instance for cases where tensorflow is not installed.
    Input = TripWire(keras_msg)


def mlp4(input_shape, n_classes, output_activation="softmax", verbose=False):
    # Z. Wang, W. Yan, T. Oates, "Time Series Classification from Scratch with Deep Neural Networks: A Strong Baseline," Int. Joint Conf. Neural Networks, 2017, pp. 1578-1585
    ip = Input(shape=input_shape)
    fc = Flatten()(ip)
    fc = Dropout(0.1)(fc)

    fc = Dense(500, activation="relu")(fc)
    fc = Dropout(0.2)(fc)

    fc = Dense(500, activation="relu")(fc)
    fc = Dropout(0.2)(fc)

    fc = Dense(500, activation="relu")(fc)
    fc = Dropout(0.3)(fc)

    out = Dense(n_classes, activation=output_activation)(fc)

    model = Model([ip], [out])
    if verbose:
        model.summary()

    return model


def cnn_lenet(input_shape, n_classes, output_activation="softmax", verbose=False):
    # Y. Lecun, L. Bottou, Y. Bengio, and P. Haffner, “Gradient-based learning applied to document recognition,” Proceedings of the IEEE, vol. 86, no. 11, pp. 2278–2324, 1998.

    ip = Input(shape=input_shape)
    conv = ip

    n_conv_layers = int(round(math.log(input_shape[0], 2)) - 3)
    if verbose:
        print("pooling layers: %d" % n_conv_layers)

    for i in range(n_conv_layers):
        conv = Conv1D(
            6 + 10 * i,
            3,
            padding="same",
            activation="relu",
            kernel_initializer="he_uniform",
        )(conv)
        conv = MaxPooling1D(pool_size=2)(conv)

    flat = Flatten()(conv)

    fc = Dense(120, activation="relu")(flat)
    fc = Dropout(0.5)(fc)

    fc = Dense(84, activation="relu")(fc)
    fc = Dropout(0.5)(fc)

    out = Dense(n_classes, activation=output_activation)(fc)

    model = Model([ip], [out])
    if verbose:
        model.summary()

    return model


def cnn_vgg(input_shape, n_classes, output_activation="softmax", verbose=False):
    # K. Simonyan and A. Zisserman, "Very deep convolutional networks for large-scale image recognition," arXiv preprint arXiv:1409.1556, 2014.

    ip = Input(shape=input_shape)
    conv = ip

    n_conv_layers = int(round(math.log(input_shape[0], 2)) - 3)
    if verbose:
        print("pooling layers: %d" % n_conv_layers)

    for i in range(n_conv_layers):
        num_filters = min(64 * 2**i, 512)
        conv = Conv1D(
            num_filters,
            3,
            padding="same",
            activation="relu",
            kernel_initializer="he_uniform",
        )(conv)
        conv = Conv1D(
            num_filters,
            3,
            padding="same",
            activation="relu",
            kernel_initializer="he_uniform",
        )(conv)
        if i > 1:
            conv = Conv1D(
                num_filters,
                3,
                padding="same",
                activation="relu",
                kernel_initializer="he_uniform",
            )(conv)
        conv = MaxPooling1D(pool_size=2)(conv)

    flat = Flatten()(conv)

    fc = Dense(4096, activation="relu")(flat)
    fc = Dropout(0.5)(fc)

    fc = Dense(4096, activation="relu")(fc)
    fc = Dropout(0.5)(fc)

    out = Dense(n_classes, activation=output_activation)(fc)

    model = Model([ip], [out])
    if verbose:
        model.summary()

    return model


def lstm1v0(input_shape, n_classes, output_activation="softmax", verbose=False):
    # Original proposal:
    # S. Hochreiter and J. Schmidhuber, “Long Short-Term Memory,” Neural Computation, vol. 9, no. 8, pp. 1735–1780, Nov. 1997.

    ip = Input(shape=input_shape)

    l2 = LSTM(512)(ip)
    out = Dense(n_classes, activation=output_activation)(l2)

    model = Model([ip], [out])
    if verbose:
        model.summary()

    return model


def lstm1(input_shape, n_classes, output_activation="softmax", verbose=False):
    # Original proposal:
    # S. Hochreiter and J. Schmidhuber, “Long Short-Term Memory,” Neural Computation, vol. 9, no. 8, pp. 1735–1780, Nov. 1997.

    # Hyperparameter choices:
    # N. Reimers and I. Gurevych, "Optimal hyperparameters for deep lstm-networks for sequence labeling tasks," arXiv, preprint arXiv:1707.06799, 2017

    ip = Input(shape=input_shape)

    l2 = LSTM(100)(ip)
    out = Dense(n_classes, activation=output_activation)(l2)

    model = Model([ip], [out])
    if verbose:
        model.summary()

    return model


def lstm2(input_shape, n_classes, output_activation="softmax", verbose=False):
    ip = Input(shape=input_shape)

    l1 = LSTM(100, return_sequences=True)(ip)
    l2 = LSTM(100)(l1)
    out = Dense(n_classes, activation=output_activation)(l2)

    model = Model([ip], [out])
    if verbose:
        model.summary()

    return model


def blstm1(input_shape, n_classes, output_activation="softmax", verbose=False):
    # Original proposal:
    # M. Schuster and K. K. Paliwal, “Bidirectional recurrent neural networks,” IEEE Transactions on Signal Processing, vol. 45, no. 11, pp. 2673–2681, 1997.

    # Hyperparameter choices:
    # N. Reimers and I. Gurevych, "Optimal hyperparameters for deep lstm-networks for sequence labeling tasks," arXiv, preprint arXiv:1707.06799, 2017
    ip = Input(shape=input_shape)

    l2 = Bidirectional(LSTM(100))(ip)
    out = Dense(n_classes, activation=output_activation)(l2)

    model = Model([ip], [out])
    if verbose:
        model.summary()

    return model


def blstm2(input_shape, n_classes, output_activation="softmax", verbose=False):
    ip = Input(shape=input_shape)

    l1 = Bidirectional(LSTM(100, return_sequences=True))(ip)
    l2 = Bidirectional(LSTM(100))(l1)
    out = Dense(n_classes, activation=output_activation)(l2)

    model = Model([ip], [out])
    if verbose:
        model.summary()

    return model


def lstm_fcn(input_shape, n_classes, output_activation="softmax", verbose=False):
    # F. Karim, S. Majumdar, H. Darabi, and S. Chen, “LSTM Fully Convolutional Networks for Time Series Classification,” IEEE Access, vol. 6, pp. 1662–1669, 2018.

    ip = Input(shape=input_shape)

    # lstm part is a 1 time step multivariate as described in Karim et al. Seems strange, but works I guess.
    lstm = Permute((2, 1))(ip)

    lstm = LSTM(128)(lstm)
    lstm = Dropout(0.8)(lstm)

    conv = Conv1D(128, 8, padding="same", kernel_initializer="he_uniform")(ip)
    conv = BatchNormalization()(conv)
    conv = Activation("relu")(conv)

    conv = Conv1D(256, 5, padding="same", kernel_initializer="he_uniform")(conv)
    conv = BatchNormalization()(conv)
    conv = Activation("relu")(conv)

    conv = Conv1D(128, 3, padding="same", kernel_initializer="he_uniform")(conv)
    conv = BatchNormalization()(conv)
    conv = Activation("relu")(conv)

    flat = GlobalAveragePooling1D()(conv)

    flat = concatenate([lstm, flat])

    out = Dense(n_classes, activation=output_activation)(flat)

    model = Model([ip], [out])
    if verbose:
        model.summary()

    return model


def cnn_resnet(input_shape, n_classes, output_activation="softmax", verbose=False):
    # I. Fawaz, G. Forestier, J. Weber, L. Idoumghar, P-A Muller, "Data augmentation using synthetic data for time series classification with deep residual networks," International Workshop on Advanced Analytics and Learning on Temporal Data ECML/PKDD, 2018

    ip = Input(shape=input_shape)
    residual = ip
    conv = ip

    for i, nb_nodes in enumerate([64, 128, 128]):
        conv = Conv1D(nb_nodes, 8, padding="same", kernel_initializer="glorot_uniform")(
            conv
        )
        conv = BatchNormalization()(conv)
        conv = Activation("relu")(conv)

        conv = Conv1D(nb_nodes, 5, padding="same", kernel_initializer="glorot_uniform")(
            conv
        )
        conv = BatchNormalization()(conv)
        conv = Activation("relu")(conv)

        conv = Conv1D(nb_nodes, 3, padding="same", kernel_initializer="glorot_uniform")(
            conv
        )
        conv = BatchNormalization()(conv)
        conv = Activation("relu")(conv)

        if i < 2:
            # expands dimensions according to Fawaz et al.
            residual = Conv1D(
                nb_nodes, 1, padding="same", kernel_initializer="glorot_uniform"
            )(residual)
        residual = BatchNormalization()(residual)
        conv = add([residual, conv])
        conv = Activation("relu")(conv)

        residual = conv

    flat = GlobalAveragePooling1D()(conv)

    out = Dense(n_classes, activation=output_activation)(flat)

    model = Model([ip], [out])
    if verbose:
        model.summary()

    return model



"""Dynamic time warping utilities for pattern-mixing based augmentation methods.

This module is copied entirely from the dtw module accompanying ref [1]_, which
is available at https://github.com/uchidalab/time_series_augmentation and is
licensed under the Apache-2.0 license. The original author of this module is
Brian Iwana.

References
----------
.. [1]  Brian Kenji Iwana and Seiichi Uchida, "An empirical survey of data
augmentation for time series classification with neural networks," PLOS ONE
16(7): e0254841. DOI: https://doi.org/10.1371/journal.pone.0254841
"""

import numpy as np
import sys

RETURN_VALUE = 0
RETURN_PATH = 1
RETURN_ALL = -1


# Core DTW
def _traceback(DTW, slope_constraint):
    i, j = np.array(DTW.shape) - 1
    p, q = [i - 1], [j - 1]

    if slope_constraint == "asymmetric":
        while i > 1:
            tb = np.argmin((DTW[i - 1, j], DTW[i - 1, j - 1], DTW[i - 1, j - 2]))

            if tb == 0:
                i = i - 1
            elif tb == 1:
                i = i - 1
                j = j - 1
            elif tb == 2:
                i = i - 1
                j = j - 2

            p.insert(0, i - 1)
            q.insert(0, j - 1)
    elif slope_constraint == "symmetric":
        while i > 1 or j > 1:
            tb = np.argmin((DTW[i - 1, j - 1], DTW[i - 1, j], DTW[i, j - 1]))

            if tb == 0:
                i = i - 1
                j = j - 1
            elif tb == 1:
                i = i - 1
            elif tb == 2:
                j = j - 1

            p.insert(0, i - 1)
            q.insert(0, j - 1)
    else:
        sys.exit("Unknown slope constraint %s" % slope_constraint)

    return (np.array(p), np.array(q))


def dtw(
    prototype,
    sample,
    return_flag=RETURN_VALUE,
    slope_constraint="asymmetric",
    window=None,
):
    """Compute the DTW of two sequences.

    Parameters
    ----------
    prototype : numpy.ndarray

    sample : numpy.ndarray
    """
    p = prototype.shape[0]
    assert p != 0, "Prototype empty!"
    s = sample.shape[0]
    assert s != 0, "Sample empty!"

    if window is None:
        window = s

    cost = np.full((p, s), np.inf)
    for i in range(p):
        start = max(0, i - window)
        end = min(s, i + window) + 1
        cost[i, start:end] = np.linalg.norm(sample[start:end] - prototype[i], axis=1)

    DTW = _cummulative_matrix(cost, slope_constraint, window)

    if return_flag == RETURN_ALL:
        return DTW[-1, -1], cost, DTW[1:, 1:], _traceback(DTW, slope_constraint)
    elif return_flag == RETURN_PATH:
        return _traceback(DTW, slope_constraint)
    else:
        return DTW[-1, -1]


def _cummulative_matrix(cost, slope_constraint, window):
    p = cost.shape[0]
    s = cost.shape[1]

    # Note: DTW is one larger than cost and the original patterns
    DTW = np.full((p + 1, s + 1), np.inf)

    DTW[0, 0] = 0.0

    if slope_constraint == "asymmetric":
        for i in range(1, p + 1):
            if i <= window + 1:
                DTW[i, 1] = cost[i - 1, 0] + min(DTW[i - 1, 0], DTW[i - 1, 1])
            for j in range(max(2, i - window), min(s, i + window) + 1):
                DTW[i, j] = cost[i - 1, j - 1] + min(
                    DTW[i - 1, j - 2], DTW[i - 1, j - 1], DTW[i - 1, j]
                )
    elif slope_constraint == "symmetric":
        for i in range(1, p + 1):
            for j in range(max(1, i - window), min(s, i + window) + 1):
                DTW[i, j] = cost[i - 1, j - 1] + min(
                    DTW[i - 1, j - 1], DTW[i, j - 1], DTW[i - 1, j]
                )
    else:
        sys.exit("Unknown slope constraint %s" % slope_constraint)

    return DTW


def shape_dtw(
    prototype,
    sample,
    return_flag=RETURN_VALUE,
    slope_constraint="asymmetric",
    window=None,
    descr_ratio=0.05,
):
    """Computes the shapeDTW of two sequences.
    :param prototype: np array [0..b]
    :param sample: np array [0..t]
    :param extended: bool
    """
    # shapeDTW
    # https://www.sciencedirect.com/science/article/pii/S0031320317303710

    p = prototype.shape[0]
    assert p != 0, "Prototype empty!"
    s = sample.shape[0]
    assert s != 0, "Sample empty!"

    if window is None:
        window = s

    p_feature_len = np.clip(np.round(p * descr_ratio), 5, 100).astype(int)
    s_feature_len = np.clip(np.round(s * descr_ratio), 5, 100).astype(int)

    # padding
    p_pad_front = (np.ceil(p_feature_len / 2.0)).astype(int)
    p_pad_back = (np.floor(p_feature_len / 2.0)).astype(int)
    s_pad_front = (np.ceil(s_feature_len / 2.0)).astype(int)
    s_pad_back = (np.floor(s_feature_len / 2.0)).astype(int)

    prototype_pad = np.pad(prototype, ((p_pad_front, p_pad_back), (0, 0)), mode="edge")
    sample_pad = np.pad(sample, ((s_pad_front, s_pad_back), (0, 0)), mode="edge")

    cost = np.full((p, s), np.inf)
    for i in range(p):
        for j in range(max(0, i - window), min(s, i + window)):
            cost[i, j] = np.linalg.norm(
                sample_pad[j : j + s_feature_len] - prototype_pad[i : i + p_feature_len]
            )

    DTW = _cummulative_matrix(cost, slope_constraint=slope_constraint, window=window)

    if return_flag == RETURN_ALL:
        return DTW[-1, -1], cost, DTW[1:, 1:], _traceback(DTW, slope_constraint)
    elif return_flag == RETURN_PATH:
        return _traceback(DTW, slope_constraint)
    else:
        return DTW[-1, -1]


# Draw helpers
def draw_graph2d(cost, DTW, path, prototype, sample):
    import matplotlib.pyplot as plt

    plt.figure(figsize=(12, 8))
    # plt.subplots_adjust(left=.02, right=.98, bottom=.001, top=.96, wspace=.05, hspace=.01)

    # cost
    plt.subplot(2, 3, 1)
    plt.imshow(cost.T, cmap=plt.cm.gray, interpolation="none", origin="lower")
    plt.plot(path[0], path[1], "y")
    plt.xlim((-0.5, cost.shape[0] - 0.5))
    plt.ylim((-0.5, cost.shape[0] - 0.5))

    # dtw
    plt.subplot(2, 3, 2)
    plt.imshow(DTW.T, cmap=plt.cm.gray, interpolation="none", origin="lower")
    plt.plot(path[0] + 1, path[1] + 1, "y")
    plt.xlim((-0.5, DTW.shape[0] - 0.5))
    plt.ylim((-0.5, DTW.shape[0] - 0.5))

    # prototype
    plt.subplot(2, 3, 4)
    plt.plot(prototype[:, 0], prototype[:, 1], "b-o")

    # connection
    plt.subplot(2, 3, 5)
    for i in range(0, path[0].shape[0]):
        plt.plot(
            [prototype[path[0][i], 0], sample[path[1][i], 0]],
            [prototype[path[0][i], 1], sample[path[1][i], 1]],
            "y-",
        )
    plt.plot(sample[:, 0], sample[:, 1], "g-o")
    plt.plot(prototype[:, 0], prototype[:, 1], "b-o")

    # sample
    plt.subplot(2, 3, 6)
    plt.plot(sample[:, 0], sample[:, 1], "g-o")

    plt.tight_layout()
    plt.show()


def draw_graph1d(cost, DTW, path, prototype, sample):
    import matplotlib.pyplot as plt

    plt.figure(figsize=(12, 8))
    # plt.subplots_adjust(left=.02, right=.98, bottom=.001, top=.96, wspace=.05, hspace=.01)
    p_steps = np.arange(prototype.shape[0])
    s_steps = np.arange(sample.shape[0])

    # cost
    plt.subplot(2, 3, 1)
    plt.imshow(cost.T, cmap=plt.cm.gray, interpolation="none", origin="lower")
    plt.plot(path[0], path[1], "y")
    plt.xlim((-0.5, cost.shape[0] - 0.5))
    plt.ylim((-0.5, cost.shape[0] - 0.5))

    # dtw
    plt.subplot(2, 3, 2)
    plt.imshow(DTW.T, cmap=plt.cm.gray, interpolation="none", origin="lower")
    plt.plot(path[0] + 1, path[1] + 1, "y")
    plt.xlim((-0.5, DTW.shape[0] - 0.5))
    plt.ylim((-0.5, DTW.shape[0] - 0.5))

    # prototype
    plt.subplot(2, 3, 4)
    plt.plot(p_steps, prototype[:, 0], "b-o")

    # connection
    plt.subplot(2, 3, 5)
    for i in range(0, path[0].shape[0]):
        plt.plot(
            [path[0][i], path[1][i]],
            [prototype[path[0][i], 0], sample[path[1][i], 0]],
            "y-",
        )
    plt.plot(p_steps, sample[:, 0], "g-o")
    plt.plot(s_steps, prototype[:, 0], "b-o")

    # sample
    plt.subplot(2, 3, 6)
    plt.plot(s_steps, sample[:, 0], "g-o")

    plt.tight_layout()
    plt.show()



"""Test functions in the h5io module

These tests copy extensively from deepdish.io to test the io functions that
were modified from the same library. Below is the entire text of deepdish's
BSD-3 license:

---

Copyright (c) 2014, Amit Group
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the {organization} nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import numpy as np
import os
import pandas as pd

from afqinsight import h5io
from contextlib import contextmanager
from tempfile import NamedTemporaryFile

try:
    from types import SimpleNamespace

    _sns = True
except ImportError:
    _sns = False


@contextmanager
def tmp_filename():
    f = NamedTemporaryFile(delete=False)
    yield f.name
    f.close()
    os.unlink(f.name)


@contextmanager
def tmp_file():
    f = NamedTemporaryFile(delete=False)
    yield f
    f.close()
    os.unlink(f.name)


def reconstruct(fn, x):
    h5io.save(fn, x)
    return h5io.load(fn)


def assert_array(fn, x):
    h5io.save(fn, x)
    x1 = h5io.load(fn)
    np.testing.assert_array_equal(x, x1)


def test_basic_data_types():
    with tmp_filename() as fn:
        x = 100
        x1 = reconstruct(fn, x)
        assert x == x1

        x = 1.23
        x1 = reconstruct(fn, x)
        assert x == x1

        x = "this is a string"
        x1 = reconstruct(fn, x)
        assert x == x1

        x = b"this is a bytearray"
        x1 = reconstruct(fn, x)
        assert x == x1

        x = None
        x1 = reconstruct(fn, x)
        assert x1 is None


def test_big_integers():
    with tmp_filename() as fn:
        x = 1239487239847234982392837423874
        x1 = reconstruct(fn, x)
        assert x == x1


def test_numpy_array():
    with tmp_filename() as fn:
        x0 = np.arange(3 * 4 * 5, dtype=np.int64).reshape((3, 4, 5))
        assert_array(fn, x0)

        x0 = x0.astype(np.float32)
        assert_array(fn, x0)

        x0 = x0.astype(np.uint8)
        assert_array(fn, x0)

        x0 = x0.astype(np.complex128)
        x0[0] = 1 + 2j
        assert_array(fn, x0)


def test_numpy_array_zero_size():
    # Arrays where one of the axes is length 0. These zero-length arrays cannot
    # be stored natively in HDF5, so we'll have to store only the shape
    with tmp_filename() as fn:
        x0 = np.arange(0, dtype=np.int64)
        assert_array(fn, x0)

        x0 = np.arange(0, dtype=np.float32).reshape((10, 20, 0))
        assert_array(fn, x0)

        x0 = np.arange(0, dtype=np.complex128).reshape((0, 5, 0))
        assert_array(fn, x0)


def test_numpy_string_array():
    with tmp_filename() as fn:
        x0 = np.array([[b"this", b"string"], [b"foo", b"bar"]])
        assert_array(fn, x0)

        x0 = np.array([["this", "string"], ["foo", "bar"]])
        assert_array(fn, x0)


def test_dictionary():
    with tmp_filename() as fn:
        d = dict(
            a=100,
            b="this is a string",
            c=np.ones(5),
            sub=dict(a=200, b="another string", c=np.random.randn(3, 4)),
        )

        d1 = reconstruct(fn, d)

        assert d["a"] == d1["a"]
        assert d["b"] == d1["b"]
        np.testing.assert_array_equal(d["c"], d1["c"])
        assert d["sub"]["a"] == d1["sub"]["a"]
        assert d["sub"]["b"] == d1["sub"]["b"]
        np.testing.assert_array_equal(d["sub"]["c"], d1["sub"]["c"])


def test_simplenamespace():
    if _sns:
        with tmp_filename() as fn:
            d = SimpleNamespace(
                a=100,
                b="this is a string",
                c=np.ones(5),
                sub=SimpleNamespace(a=200, b="another string", c=np.random.randn(3, 4)),
            )

            d1 = reconstruct(fn, d)

            assert d.a == d1.a
            assert d.b == d1.b
            np.testing.assert_array_equal(d.c, d1.c)
            assert d.sub.a == d1.sub.a
            assert d.sub.b == d1.sub.b
            np.testing.assert_array_equal(d.sub.c, d1.sub.c)


def test_softlinks_recursion():
    with tmp_filename() as fn:
        A = np.random.randn(3, 3)
        df = pd.DataFrame({"int": np.arange(3), "name": ["zero", "one", "two"]})
        AA = 4
        s = dict(A=A, B=A, c=A, d=A, f=A, g=[A, A, A], AA=AA, h=AA, df=df, df2=df)
        s["g"].append(s)
        n = reconstruct(fn, s)
        assert n["g"][0] is n["A"]
        assert (
            n["A"]
            is n["B"]
            is n["c"]
            is n["d"]
            is n["f"]
            is n["g"][0]
            is n["g"][1]
            is n["g"][2]
        )
        assert n["g"][3] is n
        assert n["AA"] == AA == n["h"]
        assert n["df"] is n["df2"]
        assert (n["df"] == df).all().all()

        # test 'sel' option on link ... need to read two vars
        # to ensure at least one is a link:
        col1 = h5io.load(fn, "/A", (slice(None), slice(1, 2)))
        assert np.all(A[:, 1] == col1.flatten())
        col1 = h5io.load(fn, "/B", (slice(None), slice(1, 2)))
        assert np.all(A[:, 1] == col1.flatten())


def test_softlinks_recursion_sns():
    if _sns:
        with tmp_filename() as fn:
            A = np.random.randn(3, 3)
            AA = 4
            s = SimpleNamespace(A=A, B=A, c=A, d=A, f=A, g=[A, A, A], AA=AA, h=AA)
            s.g.append(s)
            n = reconstruct(fn, s)
            assert n.g[0] is n.A
            assert n.A is n.B is n.c is n.d is n.f is n.g[0] is n.g[1] is n.g[2]
            assert n.g[3] is n
            assert n.AA == AA == n.h


def test_pickle_recursion():
    with tmp_filename() as fn:
        f = {4: 78}
        f["rec"] = f
        g = [23.4, f]
        h = dict(f=f, g=g)
        h2 = reconstruct(fn, h)
        assert h2["g"][0] == 23.4
        assert h2["g"][1] is h2["f"]["rec"] is h2["f"]
        assert h2["f"][4] == 78


def test_list_recursion():
    with tmp_filename() as fn:
        lst = [1, 3]
        inlst = ["inside", "list", lst]
        inlst.append(inlst)
        lst.append(lst)
        lst.append(inlst)
        lst2 = reconstruct(fn, lst)
        assert lst2[2] is lst2
        assert lst2[3][2] is lst2
        assert lst[3][2] is lst
        assert lst2[3][3] is lst2[3]
        assert lst[3][3] is lst[3]


def test_list():
    with tmp_filename() as fn:
        x = [100, "this is a string", np.ones(3), dict(foo=100)]

        x1 = reconstruct(fn, x)

        assert isinstance(x1, list)
        assert x[0] == x1[0]
        assert x[1] == x1[1]
        np.testing.assert_array_equal(x[2], x1[2])
        assert x[3]["foo"] == x1[3]["foo"]


def test_tuple():
    with tmp_filename() as fn:
        x = (100, "this is a string", np.ones(3), dict(foo=100))

        x1 = reconstruct(fn, x)

        assert isinstance(x1, tuple)
        assert x[0] == x1[0]
        assert x[1] == x1[1]
        np.testing.assert_array_equal(x[2], x1[2])
        assert x[3]["foo"] == x1[3]["foo"]


def test_sparse_matrices():
    import scipy.sparse as S

    with tmp_filename() as fn:
        x = S.lil_matrix((50, 70))
        x[34, 37] = 1
        x[34, 39] = 2.5
        x[34, 41] = -2
        x[38, 41] = -1

        x1 = reconstruct(fn, x.tocsr())
        assert x.shape == x1.shape
        np.testing.assert_array_equal(x.todense(), x1.todense())

        x1 = reconstruct(fn, x.tocsc())
        assert x.shape == x1.shape
        np.testing.assert_array_equal(x.todense(), x1.todense())

        x1 = reconstruct(fn, x.tocoo())
        assert x.shape == x1.shape
        np.testing.assert_array_equal(x.todense(), x1.todense())

        x1 = reconstruct(fn, x.todia())
        assert x.shape == x1.shape
        np.testing.assert_array_equal(x.todense(), x1.todense())

        x1 = reconstruct(fn, x.tobsr())
        assert x.shape == x1.shape
        np.testing.assert_array_equal(x.todense(), x1.todense())


def test_array_scalar():
    with tmp_filename() as fn:
        v = np.array(12.3)
        v1 = reconstruct(fn, v)
        assert v1[()] == v and isinstance(v1[()], np.float64)

        v = np.array(40, dtype=np.int8)
        v1 = reconstruct(fn, v)
        assert v1[()] == v and isinstance(v1[()], np.int8)


def test_load_group():
    with tmp_filename() as fn:
        x = dict(one=np.ones(10), two="string")
        h5io.save(fn, x)

        one = h5io.load(fn, "/one")
        np.testing.assert_array_equal(one, x["one"])
        two = h5io.load(fn, "/two")
        assert two == x["two"]

        full = h5io.load(fn, "/")
        np.testing.assert_array_equal(x["one"], full["one"])
        assert x["two"] == full["two"]


def test_load_multiple_groups():
    with tmp_filename() as fn:
        x = dict(one=np.ones(10), two="string", three=200)
        h5io.save(fn, x)

        one, three = h5io.load(fn, ["/one", "/three"])
        np.testing.assert_array_equal(one, x["one"])
        assert three == x["three"]

        three, two = h5io.load(fn, ["/three", "/two"])
        assert three == x["three"]
        assert two == x["two"]


def test_load_slice():
    with tmp_filename() as fn:
        x = np.arange(3 * 4 * 5).reshape((3, 4, 5))
        h5io.save(fn, dict(x=x))

        s = slice(None, 2)
        xs = h5io.load(fn, "/x", sel=s)
        np.testing.assert_array_equal(xs, x[s])

        s = (slice(None), slice(1, 3))
        xs = h5io.load(fn, "/x", sel=s)
        np.testing.assert_array_equal(xs, x[s])

        xs = h5io.load(fn, sel=s, unpack=True)
        np.testing.assert_array_equal(xs, x[s])

        h5io.save(fn, x)
        xs = h5io.load(fn, sel=s)
        np.testing.assert_array_equal(xs, x[s])


def test_force_pickle1():
    with tmp_filename() as fn:
        x = dict(one=dict(two=np.arange(10)), three="string")
        xf = dict(one=dict(two=x["one"]["two"]), three=x["three"])

        h5io.save(fn, xf)
        xs = h5io.load(fn)

        np.testing.assert_array_equal(x["one"]["two"], xs["one"]["two"])
        assert x["three"] == xs["three"]

        # Try direct loading one
        two = h5io.load(fn, "/one/two")
        np.testing.assert_array_equal(x["one"]["two"], two)


def test_non_string_key_dict():
    with tmp_filename() as fn:
        # These will be pickled, but it should still work
        x = {0: "zero", 1: "one", 2: "two"}
        x1 = reconstruct(fn, x)
        assert x == x1

        x = {1 + 1j: "zero", b"test": "one", (1, 2): "two"}
        x1 = reconstruct(fn, x)
        assert x == x1


def test_force_pickle2():
    with tmp_filename() as fn:
        x = {0: "zero", 1: "one", 2: "two"}
        fx = h5io.ForcePickle(x)
        d = dict(foo=x, bar=100)
        fd = dict(foo=fx, bar=100)
        d1 = reconstruct(fn, fd)
        assert d == d1


def test_pandas_dataframe():
    with tmp_filename() as fn:
        # These will be pickled, but it should still work
        df = pd.DataFrame({"int": np.arange(3), "name": ["zero", "one", "two"]})
        df1 = reconstruct(fn, df)
        assert (df == df1).all().all()


def test_pandas_series():
    rs = np.random.RandomState(1234)
    with tmp_filename() as fn:
        s = pd.Series(rs.randn(5), index=["a", "b", "c", "d", "e"])
        s1 = reconstruct(fn, s)
        assert (s == s1).all()


def test_compression_true():
    rs = np.random.RandomState(1234)
    with tmp_filename() as fn:
        x = rs.normal(size=(1000, 5))
        for comp in [None, True, "blosc", "zlib", ("zlib", 5)]:
            h5io.save(fn, x, compression=comp)
            x1 = h5io.load(fn)
            assert (x == x1).all()



"""
===============================
Harmonize HBN data using ComBat
===============================

This example loads AFQ data from the Healthy Brain Network (HBN) preprocessed
diffusion derivatives [1]_. The HBN is a landmark pediatric mental health study.
Over the course of the study, it will collect diffusion MRI data from
approximately 5,000 children and adolescents. We recently processed the
available data from over 2,000 of these subjects, and provide the tract profiles
from this dataset, which can be downloaded from AWS thanks to
[INDI](http://fcon_1000.projects.nitrc.org/).

We first load the data by using the :func:`AFQDataset.from_files` static method
and supplying AWS S3 URIs instead of local file names. We then impute missing
values and plot the mean bundle profiles by scanning site, noting that there are
substantial site differences. Lastly, we harmonize the site differences using
NeuroComBat [2]_ and plot the harmonized bundle profiles to verify that the site
differences have been removed.

.. [1]  Adam Richie-Halford, Matthew Cieslak, Lei Ai, Sendy Caffarra, Sydney
   Covitz, Alexandre R. Franco, Iliana I. Karipidis, John Kruper, Michael
   Milham, Bárbara Avelar-Pereira, Ethan Roy, Valerie J. Sydnor, Jason Yeatman,
   The Fibr Community Science Consortium, Theodore D. Satterthwaite, and Ariel
   Rokem,
   "An open, analysis-ready, and quality controlled resource for pediatric brain
   white-matter research"
   bioRxiv 2022.02.24.481303;
   doi: https://doi.org/10.1101/2022.02.24.481303

.. [2] Jean-Philippe Fortin, Drew Parker, Birkan Tunc, Takanori Watanabe, Mark A
   Elliott, Kosha Ruparel, David R Roalf, Theodore D Satterthwaite, Ruben C Gur,
   Raquel E Gur, Robert T Schultz, Ragini Verma, Russell T Shinohara.
   "Harmonization Of Multi-Site Diffusion Tensor Imaging Data"
   NeuroImage, 161, 149-170, 2017;
   doi: https://doi.org/10.1016/j.neuroimage.2017.08.047

"""
import numpy as np

from afqinsight import AFQDataset
from afqinsight.plot import plot_tract_profiles
from neurocombat_sklearn import CombatModel
from sklearn.impute import SimpleImputer
from sklearn.model_selection import train_test_split

#############################################################################
# Fetch the HBN data
# ------------------
# As a shortcut, we have incorporated a few studies  into the software. In these
# cases, a :class:`AFQDataset` class instance can be initialized using the
# :func:`AFQDataset.from_study` static method. This expects the name of one of
# the studies that are supported (see the method documentation for the list of
# these studies). By passing `"hbn"`, we request that the object download the
# HBN dataset from the AWS Open Data program where it has been stored and
# initialize the objects with the subjects and nodes information. Subjects' age
# is set as the target variable. After dropping subjects that don't have their
# age recorded, there are 1867 subjects in the dataset.


dataset = AFQDataset.from_study("hbn")
dataset.drop_target_na()
print(dataset)

#############################################################################
# Train / test split
# ------------------
#
# We can pass the :class:`AFQDataset` class instance to scikit-learn's
# :func:`train_test_split` function, just as we would with an array.

dataset_train, dataset_test = train_test_split(dataset, test_size=0.5)

##########################################################################
# Impute missing values
# ---------------------
#
# Next we impute missing values using median imputation. We fit the imputer
# using the training set and then use it to transform both the training and test
# sets.

imputer = dataset_train.model_fit(SimpleImputer(strategy="median"))
dataset_train = dataset_train.model_transform(imputer)
dataset_test = dataset_test.model_transform(imputer)

##########################################################################
# Plot average bundle profiles by scan site
# -----------------------------------------
#
# Next we plot the mean bundle profiles in the test set by scanning site. The
# :func:`plot_tract_profiles` function takes as input an :class:`AFQDataset` and
# returns matplotlib figures displaying the mean bundle profile for each bundle
# and metric, optionally grouped by a categorical or continuous variable.

site_figs = plot_tract_profiles(
    X=dataset_test,
    group_by=dataset_test.classes["scan_site_id"][dataset_test.y[:, 2].astype(int)],
    group_by_name="Site",
    figsize=(14, 14),
)

##########################################################################
# Harmonize the sites and replot
# ------------------------------
#
# We can see that there are substantial scan site differences in both the
# FA and MD profiles. Let's use neuroComBat to harmonize the site differences
# and then replot the mean bundle profiles.
#
# N.B. We use the excellent `neurocombat_sklearn
# <https://github.com/Warvito/neurocombat_sklearn>`_ package to apply ComBat to
# our data. We love this library, however, it is not fully compliant with the
# scikit-learn transformer API, so we cannot use the
# :func:`AFQDataset.model_fit_transform` method to apply this transformer to our
# dataset. No problem! We can simply copy the unharmonized dataset into a new
# variable and then overwrite the features of the new dataset with the ComBat
# output.
#
# Lastly, we replot the mean bundle profiles and confirm that ComBat did its
# job.

# Fit the ComBat transformer to the training set
combat = CombatModel()
combat.fit(
    dataset_train.X,
    dataset_train.y[:, 2][:, np.newaxis],
    dataset_train.y[:, 1][:, np.newaxis],
    dataset_train.y[:, 0][:, np.newaxis],
)

# And then transform a copy of the test set
harmonized_test = dataset_test.copy()
harmonized_test.X = combat.transform(
    dataset_test.X,
    dataset_test.y[:, 2][:, np.newaxis],
    dataset_test.y[:, 1][:, np.newaxis],
    dataset_test.y[:, 0][:, np.newaxis],
)

site_figs = plot_tract_profiles(
    X=harmonized_test,
    group_by=harmonized_test.classes["scan_site_id"][
        harmonized_test.y[:, 2].astype(int)
    ],
    group_by_name="Site",
    figsize=(14, 14),
)



"""Plot bundle profiles."""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from collections import OrderedDict
from groupyr.transform import GroupExtractor
from tqdm.auto import tqdm

from .utils import BUNDLE_MAT_2_PYTHON
from .datasets import AFQDataset


POSITIONS = OrderedDict(
    {
        "IFO_L": (0, 0),
        "IFO_R": (0, 3),
        "IFOF_L": (0, 0),
        "IFOF_R": (0, 3),
        "UNC_L": (0, 1),
        "UNC_R": (0, 2),
        "ATR_L": (1, 0),
        "ATR_R": (1, 3),
        "CST_L": (1, 1),
        "CST_R": (1, 2),
        "ARC_L": (2, 0),
        "ARC_R": (2, 3),
        "SLF_L": (2, 1),
        "SLF_R": (2, 2),
        "ILF_L": (3, 0),
        "ILF_R": (3, 3),
        "CGC_L": (3, 1),
        "CGC_R": (3, 2),
        "HCC_L": (4, 0),
        "HCC_R": (4, 3),
        "FA": (4, 1),
        "FP": (4, 2),
        "CC_ForcepsMinor": (4, 1),
        "CC_ForcepsMajor": (4, 2),
        "Orbital": (4, 0),
        "AntFrontal": (4, 1),
        "SupFrontal": (4, 2),
        "Motor": (4, 3),
        "SupParietal": (5, 0),
        "Temporal": (5, 1),
        "PostParietal": (5, 2),
        "Occipital": (5, 3),
    }
)


def plot_tract_profiles(
    X,
    groups=None,
    group_names=None,
    group_by=None,
    group_by_name=None,
    bins=None,
    quantiles=None,
    palette="colorblind",
    ci=95.0,
    subplot_positions=None,
    nrows=None,
    ncols=None,
    legend_kwargs=None,
    figsize=None,
    fig_tight_layout_kws=None,
):
    """Plot profiles for each bundle and each metric.

    Parameters
    ----------
    X : numpy.ndarray or AFQDataset class instance
        If array, this is a matrix of tractometry features with shape (n_subjects, n_features).

    groups : list of numpy.ndarray, optional
        feature indices for each feature group of ``X``.
        Must be provided if ``X`` is an array. Should not be provided if
        ``X`` is an AFQDataset.

    group_names : list of tuples
        the multi-indexed name for each group in ``groups``. Must be of same
        length as ``groups``. Must be provided if ``X`` is an array.
        Should not be provided if ``X`` is an AFQDataset

    group_by : list-like
        grouping variable that will produce different bundle profiles with
        different colors. If ``group_by`` is categorical, this function will
        produce a mean bundle profile for each category and color them
        differently. If ``group_by`` is numerical, please also provide the
        ``bins`` or ``quantiles`` parameter to convert this variable into a
        categorical variable. Default: no grouping.

    group_by_name : str
        The name of the group_by variable to be used in the plot legend.

    bins : int, sequence of scalars, or pandas.IntervalIndex
        The criteria by which to bin the ``group_by`` data.

        * int : Defines the number of equal-width bins in the range of `X`. The
          range of `X` is extended by .1% on each side to include the minimum
          and maximum values of `X`.
        * sequence of scalars : Defines the bin edges allowing for non-uniform
          width. No extension of the range of `x` is done.
        * IntervalIndex : Defines the exact bins to be used. Note that
          IntervalIndex for `bins` must be non-overlapping.

    quantiles : int or list-like of float
        Number of quantiles. 10 for deciles, 4 for quartiles, etc. Alternately
        array of quantiles, e.g. [0, .25, .5, .75, 1.] for quartiles.

    palette : string, list, dict, or matplotlib.colors.Colormap, default="colorblind"
        Method for choosing the colors to use when mapping the ``group_by``
        onto hue. String values are passed to seaborn.color_palette(). List
        or dict values imply categorical mapping, while a colormap object
        implies numeric mapping.

    ci : float, default=95.0
        Confidence interval with which to shade the bundle profiles

    subplot_positions : dict, optional
        Dictionary with keys corresponding to bundle names and values
        corresponding to the subplot positions. If not provided, this function
        will use afqinsight.plot.POSITIONS.

    nrows : int, optional
        Number of subplot rows.

    ncols : int, optional
        Number of subplot columns.

    legend_kwargs : dict, optional
        Keyword arguments to pass to the legend.

    figsize : tuple, optional
        Figure size for each figure.

    fig_tight_layout_kws : dict, optional
        Keyword arguments to pass to fig.tight_layout.

    Returns
    -------
    figs : dict
        dictionary of matplotlib figures, with keys corresponding to the
        different diffusion metrics
    """
    if isinstance(X, AFQDataset):
        if groups is not None or group_names is not None:
            raise ValueError(
                "You provided an AFQDataset class instance as `X` input and also a `groups` or `group_names` input, but these are mutually exclusive."
            )
        # Allocate the variables needed below based on the input dataset:
        group_names = X.group_names
        groups = X.groups
        X = X.X

    else:
        if groups is None or group_names is None:
            raise ValueError(
                "You provided an array input as `X` but did not provide both a `groups` and a `group_names` input. You must provide both of these for array input. "
            )

    plt_positions = subplot_positions if subplot_positions is not None else POSITIONS

    if bins is not None and quantiles is not None:
        raise ValueError(
            "You specified both bins and quantiles. These parameters are mutually exclusive."
        )

    if (bins is not None or quantiles is not None) and group_by_name is None:
        raise ValueError(
            "You must supply a group_by_name when binning using either the bins or quantiles parameter."
        )

    if group_by is None:
        if group_by_name is not None:
            raise ValueError(
                "You must supply a `group_by` value if a `group_by_name` is provided."
            )
        group_by = np.ones(X.shape[0])
        group_by_name = None

    figs = {}

    metrics = np.unique([grp[0] for grp in group_names])
    tract_names = np.unique([grp[1] for grp in group_names])

    if len(group_by.shape) == 1:
        group_by = np.copy(group_by)[:, np.newaxis]

    groups_metric = [grp for gn, grp in zip(group_names, groups) if metrics[0] in gn]

    for metric in tqdm(metrics):
        X_metric = GroupExtractor(
            select=metric, groups=groups, group_names=group_names
        ).fit_transform(X)
        group_names_metric = [gn for gn in group_names if metric in gn]

        # Create a dataframe for each bundle
        tract_stats = {}
        for tid in tract_names:
            X_select = GroupExtractor(
                select=tid, groups=groups_metric, group_names=group_names_metric
            ).fit_transform(X_metric)
            columns = [idx for idx in range(X_select.shape[1])]
            df = pd.concat(
                [
                    pd.DataFrame(X_select, columns=columns, dtype=np.float64),
                    pd.DataFrame(group_by, columns=["group_by"]),
                ],
                axis="columns",
            )
            id_vars = ["group_by"]
            hue = "group_by"  # used later in seaborn functions

            if quantiles is not None:
                df["quantile"], _bins = pd.qcut(
                    df["group_by"].astype(np.float64), quantiles, retbins=True
                )
                hue = "quantile"
                id_vars.append("quantile")

            if bins is not None:
                df["bin"], _bins = pd.cut(
                    df["group_by"].astype(np.float64), bins, retbins=True
                )
                hue = "bin"
                id_vars.append("bin")

            df = df.melt(id_vars=id_vars, var_name="nodeID", value_name=metric)
            df["nodeID"] = df["nodeID"].astype(int)
            tract_stats[tid] = df

        # Arrange the bundles into a grid
        bgcolor = "white"

        # Create the subplots
        if nrows is None:
            nrows = 5
            cc_bundles = ["PostParietal", "SupFrontal", "SupParietal", "Temporal"]
            if any([tid in tract_stats.keys() for tid in cc_bundles]):
                nrows = 6

        ncols = ncols if ncols is not None else 4

        fig, axes = plt.subplots(nrows=nrows, ncols=ncols, sharex=True, figsize=figsize)

        for tid, df_stat in tqdm(tract_stats.items()):
            bundle_id = BUNDLE_MAT_2_PYTHON.get(tid, tid)

            if "HCC_" in bundle_id or "Cingulum Hippocampus" in tid:
                continue

            ax = axes[plt_positions[bundle_id][0], plt_positions[bundle_id][1]]

            if metric == "dki_md":
                df_stat[metric] *= 1000.0

            _ = sns.lineplot(
                x="nodeID",
                y=metric,
                hue=hue,
                data=df_stat,
                ci=ci,
                palette=palette,
                ax=ax,
                linewidth=1.0,
                n_boot=500,
            )

            if plt_positions[bundle_id][0] == nrows - 1:
                _ = ax.set_xlabel("% distance along fiber bundle")

            if plt_positions[bundle_id][1] == 0 or (
                plt_positions[bundle_id][1] == 1
                and plt_positions[bundle_id][0] == 4
                and nrows == 5
                and subplot_positions is None
            ):
                _ = ax.set_ylabel(metric.lower().replace("_", " "))
            else:
                _ = ax.set(ylabel=None)

            _ = ax.tick_params(axis="both", which="major")
            _ = ax.set_facecolor(bgcolor)
            _ = ax.get_legend().remove()

            _ = ax.set_title(
                bundle_id.replace("_", "").replace("FA", "CFA").replace("FP", "CFP")
            )

        if subplot_positions is None and nrows == 5:
            _ = axes[4, 0].axis("off")
            _ = axes[4, 3].axis("off")

        handles, labels = ax.get_legend_handles_labels()

        if quantiles is not None or bins is not None:
            labels = [
                f"{group_by_name} {b[0]:.2f}-{b[1]:.2f}"
                for b in zip(_bins[:-1], _bins[1:])
            ]
        if group_by_name is not None:
            figlegend_kwargs = dict(
                facecolor="whitesmoke",
                bbox_to_anchor=(0.5, 0.02),
                loc="upper center",
                ncol=6,
            )

            if legend_kwargs is not None:
                figlegend_kwargs.update(legend_kwargs)

            leg = plt.figlegend(
                handles,
                labels,
                **figlegend_kwargs,
            )

            # set the linewidth of each legend object
            for legobj in leg.legendHandles:
                _ = legobj.set_linewidth(3.0)

        if fig_tight_layout_kws is None:
            fig_tight_layout_kws = dict(h_pad=0.5, w_pad=-0.5)

        fig.tight_layout(**fig_tight_layout_kws)

        figs[metric] = fig

    return figs



"""Test functions in the match module"""

import afqinsight.match as aim
import numpy as np
import pandas as pd
from sklearn.utils._testing import assert_array_almost_equal, assert_array_equal
from scipy.spatial.distance import mahalanobis
import pytest


test_features = np.asarray([[1, 2, 3, 4], [4, 2, 1, 3], [1, 4, 5, 6]])


ctrl_features = np.asarray(
    [
        [3, 2, 3, 4],
        [4, 5, 1, 3],
        [1, 10, 5, 6],
        [1, 2, 2, 4],
        [4, 2, 0, 3],
        [1, 4, 9, 6],
    ]
)


data = pd.DataFrame(
    {
        "eid": [34, 35, 36, 37, 38, 39, 40, 41, 42],
        "age": [12, 345, 63, 4, 5, 6, 89, 88, 10],
        "status": [1, 1, 1, 0, 0, 0, 0, 0, 0],
    }
)
for ii in range(4):
    data[f"feature_{ii}"] = [*test_features[:, ii], *ctrl_features[:, ii]]


def test_mahalonobis_dist():
    nbrs = aim._mahalonobis_dist(test_features, ctrl_features)

    v_inv = np.linalg.inv(
        np.cov(np.concatenate((test_features, ctrl_features), axis=0).T, ddof=0)
    )
    nbrs_scipy = np.zeros((test_features.shape[0], ctrl_features.shape[0]))
    for ii, test_sub in enumerate(test_features):
        for jj, ctrl_sub in enumerate(ctrl_features):
            nbrs_scipy[ii, jj] = mahalanobis(test_sub, ctrl_sub, v_inv)

    assert_array_almost_equal(nbrs, nbrs_scipy)


def test_mahalonobis_dist_match_df():
    matched_df = aim.mahalonobis_dist_match(
        data=data,
        status_col="status",
        feature_cols=["feature_0", "feature_1", "feature_2", "feature_3"],
        threshold=1,
    )

    assert_array_equal(
        matched_df.loc[matched_df.status.astype(bool)].eid.to_numpy(), [34, 35]
    )
    assert_array_equal(
        matched_df.loc[~matched_df.status.astype(bool)].eid.to_numpy(), [40, 41]
    )


def test_mahalonobis_dist_match_df_err():
    with pytest.raises(ValueError):  # no status column
        aim.mahalonobis_dist_match(data=data)
    with pytest.raises(ValueError):  # status has more than two unique values
        aim.mahalonobis_dist_match(data=data, status_col="feature_0")


def test_mahalonobis_dist_match_df_feauture_none():
    data_wo_extras = data.drop(columns=["eid", "age"])
    matched_df = aim.mahalonobis_dist_match(
        data=data_wo_extras,
        status_col="status",
        threshold=1,
    )

    assert_array_equal(
        matched_df.loc[matched_df.status.astype(bool)].feature_2.to_numpy(), [3, 1]
    )
    assert_array_equal(
        matched_df.loc[~matched_df.status.astype(bool)].feature_2.to_numpy(), [2, 0]
    )


def test_mahalonobis_dist_match():
    matched_df = aim.mahalonobis_dist_match(
        test=test_features, ctrl=ctrl_features, threshold=1
    )

    assert_array_equal(
        matched_df.loc[matched_df.status.astype(bool)].feature_2.to_numpy(), [3, 1]
    )
    assert_array_equal(
        matched_df.loc[~matched_df.status.astype(bool)].feature_2.to_numpy(), [2, 0]
    )
    assert np.sum(matched_df.status.to_numpy() == 1) == np.sum(
        matched_df.status.to_numpy() == 0
    )


def test_mahalonobis_dist_match_err():
    with pytest.raises(ValueError):  # test if nan error is raised
        test_features_with_nans = test_features.copy()
        test_features_with_nans[0, 2] = np.nan
        aim.mahalonobis_dist_match(
            test=test_features_with_nans, ctrl=ctrl_features, threshold=1
        )



import json
import numpy as np
import os.path as op
import pandas as pd
import pytest

import afqinsight as afqi
from afqinsight import AFQDataFrameMapper
from afqinsight import (
    multicol2dicts,
    multicol2sets,
    sort_features,
    beta_hat_by_groups,
    unfold_beta_hat_by_metrics,
)
from afqinsight.transform import isiterable

data_path = op.join(afqi.__path__[0], "data")
test_data_path = op.join(data_path, "test_data")


@pytest.mark.parametrize("concat_subject_session", [True, False])
@pytest.mark.parametrize("set_ses_to_int", [True, False])
def test_AFQDataFrameMapper(concat_subject_session, set_ses_to_int):
    nodes_path = op.join(test_data_path, "nodes.csv")
    nodes = pd.read_csv(nodes_path)
    if set_ses_to_int:
        nodes["sessionID"] = 1
    transformer = AFQDataFrameMapper(concat_subject_session=concat_subject_session)
    X = transformer.fit_transform(nodes)
    groups = transformer.groups_
    cols = transformer.feature_names_
    subjects = transformer.subjects_

    X_ref = np.load(op.join(test_data_path, "test_transform_x.npy"))
    groups_ref = np.load(op.join(test_data_path, "test_transform_groups.npy"))
    cols_ref = [
        tuple(item)
        for item in np.load(op.join(test_data_path, "test_transform_cols.npy"))
    ]

    assert np.allclose(X, X_ref, equal_nan=True)  # nosec
    assert np.allclose(groups, groups_ref)  # nosec
    assert cols == cols_ref  # nosec
    if concat_subject_session:
        assert set(subjects) == set(
            (nodes.subjectID + nodes.sessionID.astype(str)).unique()
        )  # nosec
    else:
        assert set(subjects) == set(nodes.subjectID.unique())  # nosec


@pytest.mark.parametrize("concat_subject_session", [True, False])
def test_AFQDataFrameMapper_mean(concat_subject_session):
    nodes_path = op.join(test_data_path, "nodes.csv")
    nodes = pd.read_csv(nodes_path)
    transformer = AFQDataFrameMapper(
        bundle_agg_func="mean", concat_subject_session=concat_subject_session
    )
    X = transformer.fit_transform(nodes)
    groups = transformer.groups_
    cols = transformer.feature_names_
    subjects = transformer.subjects_

    X_ref = (
        nodes.groupby(["subjectID", "tractID", "sessionID"])
        .agg("mean")
        .drop("nodeID", axis="columns")
        .unstack("tractID")
        .to_numpy()
    )
    groups_ref = [np.array([idx]) for idx in range(X.shape[1])]
    cols_ref = set(
        [
            tuple([item[0], item[1]])
            for item in np.load(op.join(test_data_path, "test_transform_cols.npy"))
        ]
    )

    assert np.allclose(groups, groups_ref)  # nosec
    assert set(cols) == cols_ref  # nosec
    if concat_subject_session:
        assert set(subjects) == set(
            (nodes.subjectID + nodes.sessionID).unique()
        )  # nosec
    else:
        assert set(subjects) == set(nodes.subjectID.unique())  # nosec
    assert np.allclose(X, X_ref, equal_nan=True)  # nosec


@pytest.mark.parametrize("bundle_agg_func", [None, "mean"])
def test_AFQDataFrameMapper_fit_transform(bundle_agg_func):
    nodes_path = op.join(test_data_path, "nodes.csv")
    nodes = pd.read_csv(nodes_path)
    transformer = AFQDataFrameMapper(
        bundle_agg_func=bundle_agg_func, pd_interpolate_kwargs={"method": "cubic"}
    )
    X_ft = transformer.fit_transform(nodes)
    groups_ft = transformer.groups_
    cols_ft = transformer.feature_names_
    subjects_ft = transformer.subjects_

    transformer = AFQDataFrameMapper(
        bundle_agg_func=bundle_agg_func, pd_interpolate_kwargs={"method": "cubic"}
    )
    transformer.fit(nodes)
    X_t = transformer.transform(nodes)
    groups_t = transformer.groups_
    cols_t = transformer.feature_names_
    subjects_t = transformer.subjects_

    assert np.allclose(X_ft, X_t, equal_nan=True)  # nosec
    assert np.allclose(groups_ft, groups_t)  # nosec
    assert cols_ft == cols_t  # nosec
    assert subjects_ft == subjects_t  # nosec


def test_isiterable():
    assert isiterable(range(10))  # nosec
    assert not isiterable(5)  # nosec
    assert isiterable(np.arange(10))  # nosec


def _is_lateral_bundle(bundle):
    return "left" in bundle.lower() or "right" in bundle.lower()


def _drop_left_right(bundle):
    return (
        bundle.replace("left", "")
        .replace("right", "")
        .replace("Left", "")
        .replace("Right", "")
        .strip(" ")
    )


def _relateralize_set(_set):
    lateral_list = list(filter(_is_lateral_bundle, _set))
    if lateral_list:
        return _set - set([_drop_left_right(lateral_list[0])])
    else:
        return _set


def test_multicol_utils():
    cols = [
        tuple(item)
        for item in np.load(op.join(test_data_path, "test_transform_cols.npy"))
    ]
    label_sets = multicol2sets(
        pd.MultiIndex.from_tuples(cols, names=["metric", "tractID", "nodeID"])
    )
    label_sets_ref = np.load(
        op.join(test_data_path, "test_multicol2sets_label_sets.npy"), allow_pickle=True
    )
    assert np.all(label_sets == label_sets_ref)  # nosec

    label_sets = multicol2sets(
        pd.MultiIndex.from_tuples(cols, names=["metric", "tractID", "nodeID"]),
        tract_symmetry=False,
    )
    label_sets_ref = np.array([_relateralize_set(_set) for _set in label_sets_ref])
    assert np.all(label_sets == label_sets_ref)  # nosec

    label_dicts = multicol2dicts(
        pd.MultiIndex.from_tuples(cols, names=["metric", "tractID", "nodeID"])
    )
    with open(
        op.join(test_data_path, "test_multicol2dicts_label_dicts.json"), "r"
    ) as fp:
        label_dicts_ref = json.load(fp)

    assert label_dicts == label_dicts_ref  # nosec

    label_dicts = multicol2dicts(
        pd.MultiIndex.from_tuples(cols, names=["metric", "tractID", "nodeID"]),
        tract_symmetry=False,
    )
    label_dicts_ref = [
        {key: val for key, val in _dict.items() if key != "symmetrized_tractID"}
        for _dict in label_dicts_ref
    ]
    assert label_dicts == label_dicts_ref  # nosec

    scores = np.arange(len(label_dicts))
    sorted_labels = sort_features(label_dicts, scores)
    assert sorted_labels == list(zip(reversed(label_dicts), reversed(scores)))  # nosec

    sorted_labels = sort_features(label_dicts, reversed(scores))
    assert sorted_labels == list(zip(label_dicts, reversed(scores)))  # nosec


def test_beta_hat_transforms():
    cols = [
        tuple(item)
        for item in np.load(op.join(test_data_path, "test_transform_cols.npy"))
    ]
    multicols = pd.MultiIndex.from_tuples(cols, names=["metric", "tractID", "nodeID"])
    beta_hat = np.random.rand(len(cols))
    result = beta_hat_by_groups(beta_hat, multicols)
    reshaped_beta = beta_hat.reshape(2, 3, -1)

    for bundle_idx, bundle in enumerate(result.keys()):
        for metric_idx, metric in enumerate(result[bundle].keys()):
            assert np.allclose(
                reshaped_beta[metric_idx, bundle_idx], result[bundle][metric]
            )

    result = unfold_beta_hat_by_metrics(
        beta_hat, multicols, tract_names=list(result.keys())
    )
    reshaped_beta = beta_hat.reshape(2, -1)
    for metric_idx, metric in enumerate(result.keys()):
        assert np.allclose(reshaped_beta[metric_idx], result[metric])

    beta_hat[0:100] = 0.0
    beta_hat[300:400] = 0.0
    result = beta_hat_by_groups(beta_hat, multicols, drop_zeros=True)
    reshaped_beta = beta_hat.reshape(2, 3, -1)

    for bundle_idx, bundle in enumerate(result.keys()):
        for metric_idx, metric in enumerate(result[bundle].keys()):
            assert np.allclose(
                reshaped_beta[metric_idx, bundle_idx + 1], result[bundle][metric]
            )



"""Data augmentation methods for (potentially multi-channel) one-dimensional data.

This module provides data augmentation methods for one-dimensional sequences.
The code itself borrows heavily from the code accompanying ref [1]_, which is
available at https://github.com/uchidalab/time_series_augmentation and is
licensed under the Apache-2.0 license. The code here has been modified to allow
independent multi-channel input. That is, is allows independent data
augmentation for each channel of multi-channel input data.

References
----------
.. [1]  Brian Kenji Iwana and Seiichi Uchida, "An empirical survey of data
augmentation for time series classification with neural networks," PLOS ONE
16(7): e0254841. DOI: https://doi.org/10.1371/journal.pone.0254841
"""

from .augmentation import *  # noqa: F401,F403



"""
=================================================
Classify ALS diagnosis from white matter features
=================================================

Predict ALS diagnosis from white matter features. This example fetches the ALS
classification dataset from Sarica et al [1]_. This dataset contains tractometry
features from 24 patients with ALS and 24 demographically matched control
subjects. The plots display the absolute value of the mean regression
coefficients (averaged across cross-validation splits) for the fractional
anisotropy (FA) features.

To save computational time, we take the first 10 principal components from each
feature group (i.e. from each metric-bundle combination).
For more details on this approach in a research setting, please see [2]_.

.. [1]  Alessia Sarica, et al.
   "The Corticospinal Tract Profile in AmyotrophicLateral Sclerosis"
   Human Brain Mapping, vol. 38, pp. 727-739, 2017
   DOI: 10.1002/hbm.23412

.. [2]  Adam Richie-Halford, Jason Yeatman, Noah Simon, and Ariel Rokem
   "Multidimensional analysis and detection of informative features in human brain white matter"
   PLOS Computational Biology, 2021
   DOI: 10.1371/journal.pcbi.1009136

"""

import matplotlib.pyplot as plt
import numpy as np

from afqinsight import AFQDataset
from afqinsight import make_afq_classifier_pipeline

from groupyr.decomposition import GroupPCA

from sklearn.impute import SimpleImputer
from sklearn.model_selection import cross_validate


#############################################################################
# Fetch data from Sarica et al.
# -----------------------------
# As a shortcut, we have incorporated a few studies into the software. In these
# cases, a :class:`AFQDataset` class instance can be initialized using the
# :func:`AFQDataset.from_study` static method. This expects the name of one of
# the studies that are supported (see the method documentation for the list of
# these studies). By passing `"sarica"`, we request that the software download
# the data from this study and initialize an object for us from this data.


afqdata = AFQDataset.from_study("sarica")

# Examine the data
# ----------------
# ``afqdata`` is an ``AFQDataset`` object, with properties corresponding to the tractometry features and phenotypic targets.

X = afqdata.X
y = afqdata.y.astype(float)  # SGL expects float targets
groups = afqdata.groups
feature_names = afqdata.feature_names
group_names = afqdata.group_names
subjects = afqdata.subjects

# Reduce data dimensionality
# --------------------------
# Here we reduce computation time by taking the first 10 principal components of
# each feature group and performing SGL logistic regression on those components.
# If you want to train an SGL model without group PCA, set ``do_group_pca =
# False``. This will increase the number of features by an order of magnitude
# and slow down execution time.

do_group_pca = True

if do_group_pca:
    n_components = 10

    # The next three lines retrieve the group structure of the group-wise PCA
    # and store it in ``groups_pca``. We do not use the imputer or GroupPCA transformer
    # for anything else
    imputer = SimpleImputer(strategy="median")
    gpca = GroupPCA(n_components=n_components, groups=groups)
    groups_pca = gpca.fit(imputer.fit_transform(X)).groups_out_

    transformer = GroupPCA
    transformer_kwargs = {"groups": groups, "n_components": n_components}
else:
    transformer = False
    transformer_kwargs = None


# Create the classification pipeline
# ----------------------------------
# The core computational machinery is a pipeline. These operate as scikit-learn
# compatible pipelines, so we can pass them to scikit-learn functions.
# There are many options that need to be set to configure the pipeline object.

pipe = make_afq_classifier_pipeline(
    imputer_kwargs={"strategy": "median"},  # Use median imputation
    use_cv_estimator=True,  # Automatically determine the best hyperparameters
    feature_transformer=transformer,  # See note above about group PCA
    feature_transformer_kwargs=transformer_kwargs,
    scaler="standard",  # Standard scale the features before regression
    groups=(
        groups_pca if do_group_pca else groups
    ),  # SGL will use the original feature groups or the PCA feature groups depending on the choice above
    verbose=0,  # Be quiet!
    pipeline_verbosity=False,  # No really, be quiet!
    tuning_strategy="bayes",  # Use BayesSearchCV to determine the optimal hyperparameters
    n_bayes_iter=20,  # Consider only this many points in hyperparameter space
    cv=3,  # Use three CV splits to evaluate each hyperparameter combination
    l1_ratio=[0.0, 1.0],  # Explore the entire range of ``l1_ratio``
    eps=5e-2,  # This is the ratio of the smallest to largest ``alpha`` value
    tol=1e-2,  # Set a lenient convergence tolerance just for this example
)

# Fit and cross-validate
# ----------------------
# The ``pipe`` object is a scikit-learn pipeline and can be used in other
# scikit-learn functions

scores = cross_validate(
    pipe, X, y, cv=5, return_train_score=True, return_estimator=True
)

# Display results
# ---------------

print(f"Mean train score: {np.mean(scores['train_score']):5.3f}")
print(f"Mean test score:  {np.mean(scores['test_score']):5.3f}")
print(f"Mean fit time:    {np.mean(scores['fit_time']):5.2f}s")
print(f"Mean score time:  {np.mean(scores['score_time']):5.2f}s")

mean_coefs = np.mean(
    np.abs([est.named_steps["estimate"].coef_ for est in scores["estimator"]]), axis=0
)

fig, ax = plt.subplots(1, 1, figsize=(8, 5))
_ = ax.plot(mean_coefs[:180], color="black", lw=2)
_ = ax.set_xlim(0, 180)

colors = plt.get_cmap("tab20").colors
for grp, grp_name, color in zip(groups_pca[:18], group_names, colors):
    _ = ax.axvspan(grp.min(), grp.max() + 1, color=color, alpha=0.8, label=grp_name[1])

box = ax.get_position()
_ = ax.set_position(
    [box.x0, box.y0 + box.height * 0.375, box.width, box.height * 0.625]
)

_ = ax.legend(loc="upper center", bbox_to_anchor=(0.5, -0.2), ncol=3)
_ = ax.set_ylabel(r"$\hat{\beta}$", fontsize=16)
_ = ax.set_xlabel("Group principal component", fontsize=16)
_ = ax.set_title("Group Principal Regression Coefficients (FA only)", fontsize=18)



import groupyr as gpr
import pytest

from afqinsight import make_afq_classifier_pipeline, make_afq_regressor_pipeline
from afqinsight.pipeline import make_base_afq_pipeline
from afqinsight._serial_bagging import SerialBaggingClassifier, SerialBaggingRegressor
from sklearn.base import is_classifier
from sklearn.ensemble import BaggingClassifier, BaggingRegressor
from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor
from sklearn.impute import KNNImputer, SimpleImputer
from sklearn.impute._iterative import IterativeImputer
from sklearn.preprocessing import MaxAbsScaler, MinMaxScaler
from sklearn.preprocessing import RobustScaler, StandardScaler
from sklearn.preprocessing import Normalizer, PowerTransformer, QuantileTransformer

scaler_args = [
    ("standard", StandardScaler),
    ("maxabs", MaxAbsScaler),
    ("minmax", MinMaxScaler),
    ("robust", RobustScaler),
    (Normalizer, Normalizer),
    (None, None),
]
imputer_args = [
    ("simple", SimpleImputer),
    ("knn", KNNImputer),
    (IterativeImputer, IterativeImputer),
]
power_args = [
    (True, PowerTransformer),
    (False, None),
    (QuantileTransformer, QuantileTransformer),
]
type_args = [
    (make_afq_classifier_pipeline, True, gpr.LogisticSGLCV),
    (make_afq_classifier_pipeline, False, gpr.LogisticSGL),
    (make_afq_regressor_pipeline, True, gpr.SGLCV),
    (make_afq_regressor_pipeline, False, gpr.SGL),
]
ensembler_args = [
    ("bagging", {True: BaggingClassifier, False: BaggingRegressor}),
    ("adaboost", {True: AdaBoostClassifier, False: AdaBoostRegressor}),
    ("serial-bagging", {True: SerialBaggingClassifier, False: SerialBaggingRegressor}),
    (AdaBoostClassifier, {True: AdaBoostClassifier, False: AdaBoostClassifier}),
    (None, None),
]


@pytest.mark.parametrize("scaler, ScalerStep", scaler_args)
@pytest.mark.parametrize("imputer, ImputerStep", imputer_args)
@pytest.mark.parametrize("feature_transformer, PowerStep", power_args)
@pytest.mark.parametrize("make_pipe, use_cv, EstimatorStep", type_args)
@pytest.mark.parametrize("target_transformer", [None, PowerTransformer])
@pytest.mark.parametrize("ensembler, EnsembleEstimators", ensembler_args)
def test_classifier_pipeline_steps(
    make_pipe,
    use_cv,
    EstimatorStep,
    scaler,
    ScalerStep,
    imputer,
    ImputerStep,
    feature_transformer,
    PowerStep,
    target_transformer,
    ensembler,
    EnsembleEstimators,
):
    pipeline = make_pipe(
        imputer=imputer,
        scaler=scaler,
        use_cv_estimator=use_cv,
        feature_transformer=feature_transformer,
        target_transformer=target_transformer,
        ensemble_meta_estimator=ensembler,
    )

    if scaler is not None:
        assert isinstance(pipeline.named_steps["scale"], ScalerStep)  # nosec
        assert (  # nosec
            pipeline.named_steps["scale"].get_params() == ScalerStep().get_params()
        )
    else:
        assert pipeline.named_steps["scale"] is None  # nosec

    if imputer is not None:
        assert isinstance(pipeline.named_steps["impute"], ImputerStep)  # nosec
        assert (  # nosec
            pipeline.named_steps["impute"].get_params() == ImputerStep().get_params()
        )
    else:
        assert pipeline.named_steps["impute"] is None  # nosec

    if feature_transformer:
        assert isinstance(pipeline.named_steps["feature_transform"], PowerStep)  # nosec
        assert (  # nosec
            pipeline.named_steps["feature_transform"].get_params()
            == PowerStep().get_params()
        )
    else:
        assert pipeline.named_steps["feature_transform"] is None  # nosec

    if ensembler is not None:
        EnsembleStep = EnsembleEstimators[is_classifier(EstimatorStep())]

    if target_transformer is None:
        if ensembler is None:
            assert isinstance(pipeline.named_steps["estimate"], EstimatorStep)  # nosec
            assert (  # nosec
                pipeline.named_steps["estimate"].get_params()
                == EstimatorStep().get_params()
            )
        else:
            assert isinstance(pipeline.named_steps["estimate"], EnsembleStep)  # nosec
            ensemble_params = pipeline.named_steps["estimate"].get_params()
            correct_params = EnsembleStep(base_estimator=EstimatorStep()).get_params()
            ensemble_base_est = ensemble_params.pop("base_estimator")
            correct_params.pop("base_estimator")
            assert ensemble_params == correct_params  # nosec
            assert isinstance(ensemble_base_est, EstimatorStep)  # nosec
    else:
        if ensembler is None:
            assert isinstance(  # nosec
                pipeline.named_steps["estimate"].regressor, EstimatorStep
            )
            assert (  # nosec
                pipeline.named_steps["estimate"].regressor.get_params()
                == EstimatorStep().get_params()
            )
        else:
            assert isinstance(  # nosec
                pipeline.named_steps["estimate"].regressor, EnsembleStep
            )
            ensemble_params = pipeline.named_steps["estimate"].regressor.get_params()
            correct_params = EnsembleStep(base_estimator=EstimatorStep()).get_params()
            ensemble_base_est = ensemble_params.pop("base_estimator")
            correct_params.pop("base_estimator")
            assert ensemble_params == correct_params  # nosec
            assert isinstance(ensemble_base_est, EstimatorStep)  # nosec


def test_pipeline_value_errors():
    with pytest.raises(ValueError):
        make_base_afq_pipeline(scaler="error")

    with pytest.raises(ValueError):
        make_base_afq_pipeline(scaler=object)

    with pytest.raises(ValueError):
        make_base_afq_pipeline(scaler=1729)

    with pytest.raises(ValueError):
        make_base_afq_pipeline(imputer="error")

    with pytest.raises(ValueError):
        make_base_afq_pipeline(imputer=object)

    with pytest.raises(ValueError):
        make_base_afq_pipeline(imputer=1729)

    with pytest.raises(ValueError):
        make_base_afq_pipeline(feature_transformer=object)

    with pytest.raises(ValueError):
        make_base_afq_pipeline(feature_transformer=1729)

    with pytest.raises(ValueError):
        make_base_afq_pipeline(estimator=object)

    with pytest.raises(ValueError):
        make_base_afq_pipeline(estimator=1729)

    with pytest.raises(ValueError):
        make_afq_regressor_pipeline(ensemble_meta_estimator="error")

    with pytest.raises(ValueError):
        make_afq_regressor_pipeline(ensemble_meta_estimator=object)

    with pytest.raises(ValueError):
        make_afq_regressor_pipeline(ensemble_meta_estimator=1729)


def test_base_pipeline_with_none_estimator():
    pipeline = make_base_afq_pipeline()
    assert pipeline.named_steps["estimate"] is None  # nosec


def test_base_pipeline_pass_kwargs():
    pipeline = make_base_afq_pipeline(scaler_kwargs={"with_mean": False})
    assert (  # nosec
        pipeline.named_steps["scale"].get_params()
        == StandardScaler(with_mean=False).get_params()
    )


def test_base_pipeline_pass_ensemble_kwargs():
    pipeline = make_afq_classifier_pipeline(
        ensemble_meta_estimator="bagging",
        ensemble_meta_estimator_kwargs={"n_estimators": 100},
    )
    ensemble_params = pipeline.named_steps["estimate"].get_params()
    assert ensemble_params["n_estimators"] == 100  # nosec



"""
=================================================
Use parametric statistics for group comparison
=================================================

As a contrast to the approach presented in other examples, we also present a
more "standard" statistical approach to analyze tract profile data. Here, we
will conduct a point-by-point group comparison using a linear model fit using
the standard ordinary least squares (OLS) method.

This example first fetches the ALS classification dataset from Sarica et al
[1]_. This dataset contains tractometry features from 24 patients with ALS and
24 demographically matched control subjects. It then uses the statsmodels
library to compare between the tract profiles of the two groups in one
tract (the left corticospinal tract) and in one feature (FA).

.. [1]  Alessia Sarica, et al. "The Corticospinal Tract Profile in
   AmyotrophicLateral Sclerosis" Human Brain Mapping, vol. 38, pp. 727-739, 2017
   DOI: 10.1002/hbm.23412

"""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from afqinsight import AFQDataset
from statsmodels.api import OLS
from statsmodels.stats.anova import anova_lm
from statsmodels.stats.multitest import multipletests

from sklearn.impute import SimpleImputer

#############################################################################
# Fetch data from Sarica et al.
# -----------------------------
# As a shortcut, we have incorporated a few studies into the software. In these
# cases, a :class:`AFQDataset` class instance can be initialized using the
# :func:`AFQDataset.from_study` static method. This expects the name of one of
# the studies that are supported (see the method documentation for the list of
# these studies). By passing `"sarica"`, we request that the software download
# the data from this study and initialize an object for us from this data.

afqdata = AFQDataset.from_study("sarica")

# Examine the data
# ----------------
# ``afqdata`` is an ``AFQDataset`` object, with properties corresponding to the
# tractometry features and phenotypic targets. In this case, y includes only the
# group to which the subjects belong, encoded as 0 (patients) or 1 (controls).

X = afqdata.X
y = afqdata.y
groups = afqdata.groups
feature_names = afqdata.feature_names
group_names = afqdata.group_names
subjects = afqdata.subjects

X = SimpleImputer(strategy="median").fit_transform(X)

# Filtering the data
# ----------------
# We start by filtering the data down to only the FA estimates in the left
# corticospinal tract. We can do that by creating a Pandas dataframe and then
# using the column labels that `afqdataset` has generated from the data.

data = pd.DataFrame(columns=afqdata.feature_names, data=X)
lcst_fa = data.filter(like="Left Corticospinal").filter(like="fa")

pvals = np.zeros(lcst_fa.shape[-1])
coefs = np.zeros(lcst_fa.shape[-1])
cis = np.zeros((lcst_fa.shape[-1], 2))

# Fit models
# ----------------
# Next, we fit a model for each node of the CST for the FA as a function of
# group. The initializer `OLS.from_formula` takes
# `R-style formulas <https://www.statsmodels.org/dev/example_formulas.html>`_
# as its model specification. Here, we are using the `"group"` variable as a
# categorical variable in the model. Much more complex linear models (including
# linear mixed effects model) are possible for datasets with more complex
# phenotypical data.

for ii, column in enumerate(lcst_fa.columns):
    feature, node, tract = column
    this = pd.DataFrame({"group": y, feature: lcst_fa[column]})
    model = OLS.from_formula(f"{feature} ~ C(group)", this)
    fit = model.fit()
    coefs[ii] = fit.params.loc["C(group)[T.1]"]
    cis[ii] = fit.conf_int(alpha=0.05).loc["C(group)[T.1]"].values
    pvals[ii] = anova_lm(fit, typ=2)["PR(>F)"][0]

# Correct for multiple comparison
# --------------------------------
# Because we conducted 100 comparisons, we need to correct the p-values that
# we obtained for the potential for a false discovery. There are multiple
# ways to conduct multuple comparison correction, and we will not get into
# the considerations in selecting a method here. For the example, we chose the
# Benjamini/Hochberg FDR controlling method. This returns a boolean array for
# the p-values that are rejected at a specified alpha level (after correction),
# as well as an array of the corrected p-values.

reject, pval_corrected, _, _ = multipletests(pvals, alpha=0.05, method="fdr_bh")
reject_idx = np.where(reject)

# Visualize
# ----------
# Using Matplotlib, we can visualize the results. The top figure shows the tract
# profiles of the two groups, with stars indicating the nodes at which the null
# hypothesis is rejected. The bottom figure shows the model coefficients
# together with their estimated 95% confidence interval.

fig, ax = plt.subplots(2, 1)
ax[0].plot(np.mean(lcst_fa.iloc[afqdata.y == 0], 0).values)
ax[0].plot(np.mean(lcst_fa.iloc[afqdata.y == 1], 0).values)
ax[0].plot(reject_idx, np.zeros(len(reject_idx)), "k*")

ax[1].plot(coefs)
ax[1].fill_between(range(100), cis[:, 0], cis[:, 1], alpha=0.5)
ax[1].plot(range(100), np.zeros(100), "k--")



"""Data augmentation methods for (potentially multi-channel) one-dimensional data.

This module provides data augmentation methods for one-dimensional sequences.
The code itself borrows heavily from the code accompanying ref [1]_, which is
available at https://github.com/uchidalab/time_series_augmentation and is
licensed under the Apache-2.0 license. The code here has been modified to allow
independent multi-channel input. That is, it allows independent data
augmentation for each channel of multi-channel input data.

References
----------
.. [1]  Brian Kenji Iwana and Seiichi Uchida, "An empirical survey of data
augmentation for time series classification with neural networks," PLOS ONE
16(7): e0254841. DOI: https://doi.org/10.1371/journal.pone.0254841
"""

import numpy as np
from tqdm import tqdm

from . import dtw

__all__ = [
    "jitter",
    "scaling",
    "permutation",
    "magnitude_warp",
    "time_warp",
    "window_slice",
    "window_warp",
    "spawner",
    "wdba",
    "random_guided_warp",
    "discriminative_guided_warp",
]


#    +--------------------------------------------------+
# ---| Random transformation-based augmentation methods |---
#    +--------------------------------------------------+


def jitter(x, sigma=0.03):
    """Add jitter, or noise, to 1D sequences [1]_.

    Parameters
    ----------
    x : numpy.ndarray
        Sequences with shape `(batch, time_steps, n_channels)`.

    sigma : float
        Standard deviation of the distribution to be added.

    Returns
    -------
    numpy.ndarray
        Numpy array of generated data of equal size of the input `x`.

    References
    ----------
    .. [1]  Terry T. Um, et al., "Data augmentation of wearable sensor data for
    parkinson’s disease monitoring using convolutional neural networks,"
    Proceedings of the 19th ACM International Conference on Multimodal
    Interaction, 2017, DOI: 10.1145/3136755.3136817
    """
    return x + np.random.normal(loc=0.0, scale=sigma, size=x.shape)


def scaling(x, sigma=0.1):
    """Change the global magnitude of an input 1D sequence [1]_.

    Parameters
    ----------
    x : numpy.ndarray
        Sequences with shape `(batch, time_steps, n_channels)`.

    sigma : float
        Standard deviation of the scaling constant.

    Returns
    -------
    numpy.ndarray
        Numpy array of generated data of equal size of the input `x`.

    References
    ----------
    .. [1]  Terry T. Um, et al., "Data augmentation of wearable sensor data for
    parkinson’s disease monitoring using convolutional neural networks,"
    Proceedings of the 19th ACM International Conference on Multimodal
    Interaction, 2017, DOI: 10.1145/3136755.3136817
    """
    factor = np.random.normal(loc=1.0, scale=sigma, size=(x.shape[0], x.shape[2]))
    return np.multiply(x, factor[:, np.newaxis, :])


def permutation(x, max_segments=5, seg_mode="equal"):
    """Return a random permutation of segments.

    Parameters
    ----------
    x : numpy.ndarray
        Sequences with shape `(batch, time_steps, n_channels)`.

    max_segments : int
        The maximum number of segments to use. The minimum number is 1.

    seg_mode : ["equal", "random"]
        If "equal", use equal sized permutation segments. If "random", use
        randomly sized segments.

    Returns
    -------
    numpy.ndarray
        Numpy array of generated data of equal size of the input `x`.

    References
    ----------
    .. [1]  Terry T. Um, et al., "Data augmentation of wearable sensor data for
    parkinson’s disease monitoring using convolutional neural networks,"
    Proceedings of the 19th ACM International Conference on Multimodal
    Interaction, 2017, DOI: 10.1145/3136755.3136817
    """
    orig_steps = np.arange(x.shape[1])

    num_segs = np.random.randint(1, max_segments, size=(x.shape[0]))

    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        if num_segs[i] > 1:
            if seg_mode == "random":
                split_points = np.random.choice(
                    x.shape[1] - 2, num_segs[i] - 1, replace=False
                )
                split_points.sort()
                splits = np.split(orig_steps, split_points)
            else:
                splits = np.array_split(orig_steps, num_segs[i])
            warp = np.concatenate(np.random.permutation(splits)).ravel()
            ret[i] = pat[warp]
        else:
            ret[i] = pat
    return ret


def magnitude_warp(x, sigma=0.2, knot=4):
    """Warp the magnitude of the input sequence.

    Scale the magnitude of each time series by a curve created by cubic spline
    with a set number of knots at random magnitudes [1]_.

    Parameters
    ----------
    x : numpy.ndarray
        Sequences with shape `(batch, time_steps, n_channels)`.

    sigma : float
        Standard devation of the random magnitudes.

    knot : int
        Number of knots, i.e. hills/valleys in the spline.

    Returns
    -------
    numpy.ndarray
        Numpy array of generated data of equal size of the input `x`.

    References
    ----------
    .. [1]  Terry T. Um, et al., "Data augmentation of wearable sensor data for
    parkinson’s disease monitoring using convolutional neural networks,"
    Proceedings of the 19th ACM International Conference on Multimodal
    Interaction, 2017, DOI: 10.1145/3136755.3136817
    """
    from scipy.interpolate import CubicSpline

    orig_steps = np.arange(x.shape[1])

    random_warps = np.random.normal(
        loc=1.0, scale=sigma, size=(x.shape[0], knot + 2, x.shape[2])
    )
    warp_steps = (
        np.ones((x.shape[2], 1)) * (np.linspace(0, x.shape[1] - 1.0, num=knot + 2))
    ).T
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        warper = np.array(
            [
                CubicSpline(warp_steps[:, dim], random_warps[i, :, dim])(orig_steps)
                for dim in range(x.shape[2])
            ]
        ).T
        ret[i] = pat * warper

    return ret


def time_warp(x, sigma=0.2, knot=4):
    """Return a random smooth time warping of the input sequences.

    This function warps the time steps based on a smooth curve defined by a cubic spline.
    See refs [1]_ and [2]_ for further details.

    Parameters
    ----------
    x : numpy.ndarray
        Sequences with shape `(batch, time_steps, n_channels)`.

    sigma : float
        Standard devation of the random magnitudes of the warping path.

    knot : int
        Number of knots, i.e. hills/valleys in the warping path.

    Returns
    -------
    numpy.ndarray
        Numpy array of generated data of equal size of the input `x`.

    References
    ----------
    .. [1]  Terry T. Um, et al., "Data augmentation of wearable sensor data for
    parkinson’s disease monitoring using convolutional neural networks,"
    Proceedings of the 19th ACM International Conference on Multimodal
    Interaction, 2017, DOI: 10.1145/3136755.3136817

    .. [2]  Brian Kenji Iwana and Seiichi Uchida, "An empirical survey of data
    augmentation for time series classification with neural networks," PLOS ONE
    16(7): e0254841. DOI: https://doi.org/10.1371/journal.pone.0254841

    """
    from scipy.interpolate import CubicSpline

    orig_steps = np.arange(x.shape[1])

    random_warps = np.random.normal(
        loc=1.0, scale=sigma, size=(x.shape[0], knot + 2, x.shape[2])
    )
    warp_steps = (
        np.ones((x.shape[2], 1)) * (np.linspace(0, x.shape[1] - 1.0, num=knot + 2))
    ).T

    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            time_warp = CubicSpline(
                warp_steps[:, dim], warp_steps[:, dim] * random_warps[i, :, dim]
            )(orig_steps)
            scale = (x.shape[1] - 1) / time_warp[-1]
            ret[i, :, dim] = np.interp(
                orig_steps, np.clip(scale * time_warp, 0, x.shape[1] - 1), pat[:, dim]
            ).T
    return ret


def window_slice(x, reduce_ratio=0.9, warp_channels_independently=False):
    """Crop the input sequence by the `reduce_ratio`.

    This function will randomly crop from both the left and right to fulfill the
    ``reduce_ratio`` requirement. The result is then "stretched" to fit on the
    original time domain. See refs [1]_ and [2]_ for further details.

    Parameters
    ----------
    x : numpy.ndarray
        Sequences with shape `(batch, time_steps, n_channels)`.

    reduce_ratio : float
        The amount of input sequence that should remain. A value of 1.0
        corresponds to no cropping.

    warp_channels_independently : bool, default=False
        If True, use independent warping windows and scales for each channel.
        If False, each channel will be warped in the same way.

    Returns
    -------
    numpy.ndarray
        Numpy array of generated data of equal size of the input `x`.

    References
    ----------
    .. [1]  A. Le Guennec, S. Malinowski, R. Tavenard, "Data Augmentation for
    Time Series Classification using Convolutional Neural Networks," in
    ECML/PKDD Workshop on Advanced Analytics and Learning on Temporal Data,
    2016. HAL Id: halshs-01357973,
    URL: https://halshs.archives-ouvertes.fr/halshs-01357973.

    .. [2]  Brian Kenji Iwana and Seiichi Uchida, "An empirical survey of data
    augmentation for time series classification with neural networks," PLOS ONE
    16(7): e0254841. DOI: https://doi.org/10.1371/journal.pone.0254841

    """
    if warp_channels_independently:
        shape = (x.shape[0], x.shape[2])
    else:
        shape = x.shape[0]

    target_len = np.ceil(reduce_ratio * x.shape[1]).astype(int)
    if target_len >= x.shape[1]:
        return x
    starts = np.random.randint(
        low=0, high=x.shape[1] - target_len, size=(shape)
    ).astype(int)
    ends = (target_len + starts).astype(int)

    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            if warp_channels_independently:
                start = starts[i, dim]
                end = ends[i, dim]
            else:
                start = starts[i]
                end = ends[i]

            ret[i, :, dim] = np.interp(
                np.linspace(0, target_len, num=x.shape[1]),
                np.arange(target_len),
                pat[start:end, dim],
            ).T
    return ret


def window_warp(
    x, window_ratio=0.1, scales=[0.5, 2.0], warp_channels_independently=False
):
    """Randomly warp a time window by scales.

    Window warping takes a random window of the input sequence and stretches or
    contracts it. See refs [1]_ and [2]_. The contraction scale is chosen randomly from the ``scales``
    parameter.

    Parameters
    ----------
    x : numpy.ndarray
        Sequences with shape `(batch, time_steps, n_channels)`.

    window_ratio : float
        Ratio of the window to the full time series.

    scales : int
        A list ratios to warp the window by.

    warp_channels_independently : bool, default=False
        If True, use independent warping windows and scales for each channel.
        If False, each channel will be warped in the same way.

    Returns
    -------
    numpy.ndarray
        Numpy array of generated data of equal size of the input `x`.

    References
    ----------
    .. [1]  A. Le Guennec, S. Malinowski, R. Tavenard, "Data Augmentation for
    Time Series Classification using Convolutional Neural Networks," in
    ECML/PKDD Workshop on Advanced Analytics and Learning on Temporal Data,
    2016. HAL Id: halshs-01357973,
    URL: https://halshs.archives-ouvertes.fr/halshs-01357973.

    .. [2]  Brian Kenji Iwana and Seiichi Uchida, "An empirical survey of data
    augmentation for time series classification with neural networks," PLOS ONE
    16(7): e0254841. DOI: https://doi.org/10.1371/journal.pone.0254841

    """
    if warp_channels_independently:
        shape = (x.shape[0], x.shape[2])
    else:
        shape = x.shape[0]

    warp_scales = np.random.choice(scales, shape)
    warp_size = np.ceil(window_ratio * x.shape[1]).astype(int)
    window_steps = np.arange(warp_size)

    window_starts = np.random.randint(
        low=1, high=x.shape[1] - warp_size - 1, size=(shape)
    ).astype(int)
    window_ends = (window_starts + warp_size).astype(int)

    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            if warp_channels_independently:
                w_start = window_starts[i, dim]
                w_scale = warp_scales[i, dim]
                w_end = window_ends[i, dim]
            else:
                w_start = window_starts[i]
                w_scale = warp_scales[i]
                w_end = window_ends[i]

            start_seg = pat[:w_start, dim]
            window_seg = np.interp(
                np.linspace(0, warp_size - 1, num=int(warp_size * w_scale)),
                window_steps,
                pat[w_start:w_end, dim],
            )
            end_seg = pat[w_end:, dim]
            warped = np.concatenate((start_seg, window_seg, end_seg))
            ret[i, :, dim] = np.interp(
                np.arange(x.shape[1]),
                np.linspace(0, x.shape[1] - 1.0, num=warped.size),
                warped,
            ).T

    return ret


#    +-------------------------------------------+
# ---| Pattern mixing-based augmentation methods |---
#    +-------------------------------------------+


def spawner(x, labels, sigma=0.05, verbose=0):
    """Augment using SuboPtimAl Warped time-series geNEratoR (SPAWNER).

    From ref [2]_:
    "SuboPtimAl Warped time series geNEratoR (SPAWNER) [1]_ was introduced...
    as a method of generating patterns through a novel method called suboptimal
    time warping. Suboptimal time warping uses the warping ability of [dynamic
    time warping] DTW but adding an additional constraint that forces the
    warping path through a random point. By using the suboptimal time warping,
    SPAWNER is able to create an almost unlimited number of new time series by
    averaging aligned patterns."

    Parameters
    ----------
    x : numpy.ndarray
        Sequences with shape `(batch, time_steps, n_channels)`.

    labels : numpy.ndarray, 2D or 3D
        Either list of integers or one hot encoding of the labels.

    sigma : float
        Standard deviation of the jittering.

    verbose : int
        Verbosity flag. `1` prints out a DTW matrix. `0` shows nothing.

    Returns
    -------
    numpy.ndarray
        Numpy array of generated data of equal size of the input `x`.

    References
    ----------
    .. [1]  K. Kamycki, T. Kapuscinski, M. Oszust, "Data Augmentation with
    Suboptimal Warping for Time-Series Classification," Sensors, vol. 20, no. 1,
    2020. URL: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6983028/

    .. [2]  Brian Kenji Iwana and Seiichi Uchida, "An empirical survey of data
    augmentation for time series classification with neural networks," PLOS ONE
    16(7): e0254841. DOI: https://doi.org/10.1371/journal.pone.0254841
    """
    random_points = np.random.randint(low=1, high=x.shape[1] - 1, size=x.shape[0])
    window = np.ceil(x.shape[1] / 10.0).astype(int)
    orig_steps = np.arange(x.shape[1])
    _labels = np.argmax(labels, axis=1) if labels.ndim > 1 else labels

    ret = np.zeros_like(x)
    for i, pat in enumerate(tqdm(x)):
        # guarentees that same one isnt selected
        choices = np.delete(np.arange(x.shape[0]), i)
        # remove ones of different classes
        choices = np.where(_labels[choices] == _labels[i])[0]
        if choices.size > 0:
            random_sample = x[np.random.choice(choices)]
            # SPAWNER splits the path into two randomly
            path1 = dtw.dtw(
                pat[: random_points[i]],
                random_sample[: random_points[i]],
                dtw.RETURN_PATH,
                slope_constraint="symmetric",
                window=window,
            )
            path2 = dtw.dtw(
                pat[random_points[i] :],
                random_sample[random_points[i] :],
                dtw.RETURN_PATH,
                slope_constraint="symmetric",
                window=window,
            )
            combined = np.concatenate(
                (np.vstack(path1), np.vstack(path2 + random_points[i])), axis=1
            )
            if verbose:
                print(random_points[i])
                _, cost, DTW_map, path = dtw.dtw(
                    pat,
                    random_sample,
                    return_flag=dtw.RETURN_ALL,
                    slope_constraint="symmetric",
                    window=window,
                )
                dtw.draw_graph1d(cost, DTW_map, path, pat, random_sample)
                dtw.draw_graph1d(cost, DTW_map, combined, pat, random_sample)
            mean = np.mean([pat[combined[0]], random_sample[combined[1]]], axis=0)
            for dim in range(x.shape[2]):
                ret[i, :, dim] = np.interp(
                    orig_steps,
                    np.linspace(0, x.shape[1] - 1.0, num=mean.shape[0]),
                    mean[:, dim],
                ).T
        else:
            print(
                "There is only one pattern of class %d, skipping pattern average"
                % _labels[i]
            )
            ret[i, :] = pat
    return jitter(ret, sigma=sigma)


def wdba(x, labels, batch_size=6, slope_constraint="symmetric", use_window=True):
    """Weighted Dynamic Time Warping Barycenter Averaging (wDBA).

    Uses the Average Selected with Distance (ASD) version of DBA from ref [1]_.

    A summary from ref [2]_:
    "DTW Barycentric Averaging (DBA) [3]_ is a method of averaging multiple
    discrete time series by finding the center of time aligned elements. It does
    this through an iterative process of adding sample patterns that are time
    aligned by DTW to a cumulative centroid pattern. The advantage of using DBA
    over linear averaging is that the underlying pattern is preserved, whereas
    linear averaging might smooth features (for example, linear averaging time
    series that are just offset in time would lose distinct features).

    For data augmentation, Forestier et al. [1]_ proposed a weighted version of
    DBA (wDBA). They propose three weighting schemes, Average All (AA), Average
    Selected (AS), and Average Selected with Distance (ASD). AA weights all of
    the time series in a class input into wDBA by a flat Dirichlet distribution.
    AS selects a reference time series and weights two of the five nearest
    neighbors by a large constant amount and all the rest by a small constant
    amount. ASD is similar to AS except that it weights based on the distance to
    the reference.
    "

    References
    ----------
    .. [1]  G. Forestier, F. Petitjean, H. A. Dau, G. I. Webb and E. Keogh,
    "Generating Synthetic Time Series to Augment Sparse Datasets," 2017 IEEE
    International Conference on Data Mining (ICDM), 2017, pp. 865-870, DOI:
    10.1109/ICDM.2017.106.

    .. [2]  Brian Kenji Iwana and Seiichi Uchida, "An empirical survey of data
    augmentation for time series classification with neural networks," PLOS ONE
    16(7): e0254841. DOI: https://doi.org/10.1371/journal.pone.0254841

    .. [3]  Petitjean F, Ketterlin A, Gançarski P. A global averaging method for
    dynamic time warping, with applications to clustering. Pattern Recogn.
    2011;44(3):678–693. DOI: https://doi.org/10.1016/j.patcog.2010.09.013
    """
    if use_window:
        window = np.ceil(x.shape[1] / 10.0).astype(int)
    else:
        window = None
    _labels = np.argmax(labels, axis=1) if labels.ndim > 1 else labels

    ret = np.zeros_like(x)
    for i in tqdm(range(ret.shape[0])):
        # get the same class as i
        choices = np.where(_labels == _labels[i])[0]
        if choices.size > 0:
            # pick random intra-class pattern
            k = min(choices.size, batch_size)
            random_prototypes = x[np.random.choice(choices, k, replace=False)]

            # calculate dtw between all
            dtw_matrix = np.zeros((k, k))
            for p, prototype in enumerate(random_prototypes):
                for s, sample in enumerate(random_prototypes):
                    if p == s:
                        dtw_matrix[p, s] = 0.0
                    else:
                        dtw_matrix[p, s] = dtw.dtw(
                            prototype,
                            sample,
                            dtw.RETURN_VALUE,
                            slope_constraint=slope_constraint,
                            window=window,
                        )

            # get medoid
            medoid_id = np.argsort(np.sum(dtw_matrix, axis=1))[0]
            nearest_order = np.argsort(dtw_matrix[medoid_id])
            medoid_pattern = random_prototypes[medoid_id]

            # start weighted DBA
            average_pattern = np.zeros_like(medoid_pattern)
            weighted_sums = np.zeros((medoid_pattern.shape[0]))
            for nid in nearest_order:
                if nid == medoid_id or dtw_matrix[medoid_id, nearest_order[1]] == 0.0:
                    average_pattern += medoid_pattern
                    weighted_sums += np.ones_like(weighted_sums)
                else:
                    path = dtw.dtw(
                        medoid_pattern,
                        random_prototypes[nid],
                        dtw.RETURN_PATH,
                        slope_constraint=slope_constraint,
                        window=window,
                    )
                    dtw_value = dtw_matrix[medoid_id, nid]
                    warped = random_prototypes[nid, path[1]]
                    weight = np.exp(
                        np.log(0.5)
                        * dtw_value
                        / dtw_matrix[medoid_id, nearest_order[1]]
                    )
                    average_pattern[path[0]] += weight * warped
                    weighted_sums[path[0]] += weight

            ret[i, :] = average_pattern / weighted_sums[:, np.newaxis]
        else:
            print(
                "There is only one pattern of class %d, skipping pattern average"
                % _labels[i]
            )
            ret[i, :] = x[i]
    return ret


def random_guided_warp(
    x, labels, slope_constraint="symmetric", use_window=True, dtw_type="normal"
):
    """Random guided warping based on ref[4]_.

    From ref [3]_:
    "Guided warping [1]_ combines time series by time warping a reference
    pattern by a teacher pattern using Dynamic Time Warping (DTW) [2]_. DTW is a
    method of measuring the distance between time series using elastic element
    matching found by dynamic programming. Guided warping uses the dynamic
    alignment function of DTW to warp the elements of a reference pattern to the
    elements of a teacher pattern. In this way, the reference pattern is set to
    the time steps of the teacher pattern. This is different from averaging in
    that the mixing happens only in the time domain. There are two variants,
    Random Guided Warping (RGW) which uses a random intra-class teacher and
    Discriminative Guided Warping (RGW) which uses a directed discriminative
    teacher [4]_."

    Parameters
    ----------
    x : numpy.ndarray
        Sequences with shape `(batch, time_steps, n_channels)`.

    labels : numpy.ndarray, 2D or 3D
        Either list of integers or one hot encoding of the labels.

    slope_constraint : ["symmetric", "asymmetric"], default="symmetric"
        Slope constraint for DTW.

    use_window : bool, default=True
        If True, use a 10% boundary constraint window for DTW.

    dtw_type : ["normal", "shape"], default="normal"
        Use DTW (`normal`) or shapeDTW (`shape`).

    Returns
    -------
    numpy.ndarray
        Numpy array of generated data of equal size of the input `x`.

    References
    ----------
    .. [1]  K. Kamycki, T. Kapuscinski, M. Oszust, "Data Augmentation with
    Suboptimal Warping for Time-Series Classification," Sensors, vol. 20, no. 1,
    2020. URL: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6983028/

    .. [2]  H. Sakoe and S. Chiba, "Dynamic programming algorithm optimization for
    spoken word recognition," in IEEE Transactions on Acoustics, Speech, and
    Signal Processing, vol. 26, no. 1, pp. 43-49, February 1978, DOI:
    10.1109/TASSP.1978.1163055.

    .. [3]  Brian Kenji Iwana and Seiichi Uchida, "An empirical survey of data
    augmentation for time series classification with neural networks," PLOS ONE
    16(7): e0254841. DOI: https://doi.org/10.1371/journal.pone.0254841

    .. [4]  B. K. Iwana, S. Uchida, "Time Series Data Augmentation for Neural
    Networks by Time Warping with a Discriminative Teacher," arXiv, 2020. URL:
    https://arxiv.org/abs/2004.08780v1
    """
    if use_window:
        window = np.ceil(x.shape[1] / 10.0).astype(int)
    else:
        window = None
    orig_steps = np.arange(x.shape[1])
    _labels = np.argmax(labels, axis=1) if labels.ndim > 1 else labels

    ret = np.zeros_like(x)
    for i, pat in enumerate(tqdm(x)):
        # guarentees that same one isnt selected
        choices = np.delete(np.arange(x.shape[0]), i)
        # remove ones of different classes
        choices = np.where(_labels[choices] == _labels[i])[0]
        if choices.size > 0:
            # pick random intra-class pattern
            random_prototype = x[np.random.choice(choices)]

            if dtw_type == "shape":
                path = dtw.shape_dtw(
                    random_prototype,
                    pat,
                    dtw.RETURN_PATH,
                    slope_constraint=slope_constraint,
                    window=window,
                )
            else:
                path = dtw.dtw(
                    random_prototype,
                    pat,
                    dtw.RETURN_PATH,
                    slope_constraint=slope_constraint,
                    window=window,
                )

            # Time warp
            warped = pat[path[1]]
            for dim in range(x.shape[2]):
                ret[i, :, dim] = np.interp(
                    orig_steps,
                    np.linspace(0, x.shape[1] - 1.0, num=warped.shape[0]),
                    warped[:, dim],
                ).T
        else:
            print(
                "There is only one pattern of class %d, skipping timewarping"
                % _labels[i]
            )
            ret[i, :] = pat
    return ret


def discriminative_guided_warp(
    x,
    labels,
    batch_size=6,
    slope_constraint="symmetric",
    use_window=True,
    dtw_type="normal",
    use_variable_slice=True,
):
    """Discriminative guided warping based on ref[4]_.

    From ref [3]_:
    "Guided warping [1]_ combines time series by time warping a reference
    pattern by a teacher pattern using Dynamic Time Warping (DTW) [2]_. DTW is a
    method of measuring the distance between time series using elastic element
    matching found by dynamic programming. Guided warping uses the dynamic
    alignment function of DTW to warp the elements of a reference pattern to the
    elements of a teacher pattern. In this way, the reference pattern is set to
    the time steps of the teacher pattern. This is different from averaging in
    that the mixing happens only in the time domain. There are two variants,
    Random Guided Warping (RGW) which uses a random intra-class teacher and
    Discriminative Guided Warping (RGW) which uses a directed discriminative
    teacher [4]_."

    Parameters
    ----------
    x : numpy.ndarray
        Sequences with shape `(batch, time_steps, n_channels)`.

    labels : numpy.ndarray, 2D or 3D
        Either list of integers or one hot encoding of the labels.

    slope_constraint : ["symmetric", "asymmetric"], default="symmetric"
        Slope constraint for DTW.

    use_window : bool, default=True
        If True, use a 10% boundary constraint window for DTW.

    dtw_type : ["normal", "shape"], default="normal"
        Use DTW (`normal`) or shapeDTW (`shape`).

    Returns
    -------
    numpy.ndarray
        Numpy array of generated data of equal size of the input `x`.

    References
    ----------
    .. [1]  K. Kamycki, T. Kapuscinski, M. Oszust, "Data Augmentation with
    Suboptimal Warping for Time-Series Classification," Sensors, vol. 20, no. 1,
    2020. URL: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6983028/

    .. [2]  H. Sakoe and S. Chiba, "Dynamic programming algorithm optimization for
    spoken word recognition," in IEEE Transactions on Acoustics, Speech, and
    Signal Processing, vol. 26, no. 1, pp. 43-49, February 1978, DOI:
    10.1109/TASSP.1978.1163055.

    .. [3]  Brian Kenji Iwana and Seiichi Uchida, "An empirical survey of data
    augmentation for time series classification with neural networks," PLOS ONE
    16(7): e0254841. DOI: https://doi.org/10.1371/journal.pone.0254841

    .. [4]  B. K. Iwana, S. Uchida, "Time Series Data Augmentation for Neural
    Networks by Time Warping with a Discriminative Teacher," arXiv, 2020. URL:
    https://arxiv.org/abs/2004.08780v1
    """
    if use_window:
        window = np.ceil(x.shape[1] / 10.0).astype(int)
    else:
        window = None
    orig_steps = np.arange(x.shape[1])
    _labels = np.argmax(labels, axis=1) if labels.ndim > 1 else labels

    positive_batch = np.ceil(batch_size / 2).astype(int)
    negative_batch = np.floor(batch_size / 2).astype(int)

    ret = np.zeros_like(x)
    warp_amount = np.zeros(x.shape[0])
    for i, pat in enumerate(tqdm(x)):
        # guarentees that same one isnt selected
        choices = np.delete(np.arange(x.shape[0]), i)

        # remove ones of different classes
        positive = np.where(_labels[choices] == _labels[i])[0]
        negative = np.where(_labels[choices] != _labels[i])[0]

        if positive.size > 0 and negative.size > 0:
            pos_k = min(positive.size, positive_batch)
            neg_k = min(negative.size, negative_batch)
            positive_prototypes = x[np.random.choice(positive, pos_k, replace=False)]
            negative_prototypes = x[np.random.choice(negative, neg_k, replace=False)]

            # vector embedding and nearest prototype in one
            pos_aves = np.zeros((pos_k))
            neg_aves = np.zeros((pos_k))
            if dtw_type == "shape":
                for p, pos_prot in enumerate(positive_prototypes):
                    for ps, pos_samp in enumerate(positive_prototypes):
                        if p != ps:
                            pos_aves[p] += (1.0 / (pos_k - 1.0)) * dtw.shape_dtw(
                                pos_prot,
                                pos_samp,
                                dtw.RETURN_VALUE,
                                slope_constraint=slope_constraint,
                                window=window,
                            )
                    for ns, neg_samp in enumerate(negative_prototypes):
                        neg_aves[p] += (1.0 / neg_k) * dtw.shape_dtw(
                            pos_prot,
                            neg_samp,
                            dtw.RETURN_VALUE,
                            slope_constraint=slope_constraint,
                            window=window,
                        )
                selected_id = np.argmax(neg_aves - pos_aves)
                path = dtw.shape_dtw(
                    positive_prototypes[selected_id],
                    pat,
                    dtw.RETURN_PATH,
                    slope_constraint=slope_constraint,
                    window=window,
                )
            else:
                for p, pos_prot in enumerate(positive_prototypes):
                    for ps, pos_samp in enumerate(positive_prototypes):
                        if p != ps:
                            pos_aves[p] += (1.0 / (pos_k - 1.0)) * dtw.dtw(
                                pos_prot,
                                pos_samp,
                                dtw.RETURN_VALUE,
                                slope_constraint=slope_constraint,
                                window=window,
                            )
                    for ns, neg_samp in enumerate(negative_prototypes):
                        neg_aves[p] += (1.0 / neg_k) * dtw.dtw(
                            pos_prot,
                            neg_samp,
                            dtw.RETURN_VALUE,
                            slope_constraint=slope_constraint,
                            window=window,
                        )
                selected_id = np.argmax(neg_aves - pos_aves)
                path = dtw.dtw(
                    positive_prototypes[selected_id],
                    pat,
                    dtw.RETURN_PATH,
                    slope_constraint=slope_constraint,
                    window=window,
                )

            # Time warp
            warped = pat[path[1]]
            warp_path_interp = np.interp(
                orig_steps,
                np.linspace(0, x.shape[1] - 1.0, num=warped.shape[0]),
                path[1],
            )
            warp_amount[i] = np.sum(np.abs(orig_steps - warp_path_interp))
            for dim in range(x.shape[2]):
                ret[i, :, dim] = np.interp(
                    orig_steps,
                    np.linspace(0, x.shape[1] - 1.0, num=warped.shape[0]),
                    warped[:, dim],
                ).T
        else:
            print("There is only one pattern of class %d" % _labels[i])
            ret[i, :] = pat
            warp_amount[i] = 0.0
    if use_variable_slice:
        max_warp = np.max(warp_amount)
        if max_warp == 0:
            # unchanged
            ret = window_slice(ret, reduce_ratio=0.9)
        else:
            for i, pat in enumerate(ret):
                # Variable Sllicing
                ret[i] = window_slice(
                    pat[np.newaxis, :, :],
                    reduce_ratio=0.9 + 0.1 * warp_amount[i] / max_warp,
                )[0]
    return ret



"""
======================================
Predict age from white matter features
======================================

Predict subject age from white matter features. This example fetches the
Weston-Havens dataset described in Yeatman et al [1]_. This dataset contains
tractometry features from 77 subjects ages 6-50. The plots display the absolute
value of the mean regression coefficients (averaged across cross-validation
splits) for the mean diffusivity (MD) features.

Predictive performance for this example is quite poor. In a research setting,
one might have to ensemble a number of SGL estimators together and conduct a
more thorough search of the hyperparameter space. For more details, please see
[2]_.

.. [1]  Jason D. Yeatman, Brian A. Wandell, & Aviv A. Mezer, "Lifespan
    maturation and degeneration of human brain white matter" Nature
    Communications, vol. 5:1, pp. 4932, 2014 DOI: 10.1038/ncomms5932

.. [2]  Adam Richie-Halford, Jason Yeatman, Noah Simon, and Ariel Rokem
   "Multidimensional analysis and detection of informative features in human
   brain white matter" PLOS Computational Biology, 2021 DOI:
   10.1371/journal.pcbi.1009136

"""
import matplotlib.pyplot as plt
import numpy as np
import os.path as op

from afqinsight.datasets import download_weston_havens, load_afq_data
from afqinsight import make_afq_regressor_pipeline

from sklearn.model_selection import cross_validate

##########################################################################
# Fetch example data
# ------------------
#
# The :func:`download_weston_havens` function download the data used in this
# example and places it in the `~/.cache/afq-insight/weston_havens` directory.
# If the directory does not exist, it is created. The data follows the format
# expected by the :func:`load_afq_data` function: a file called `nodes.csv` that
# contains AFQ tract profiles and a file called `subjects.csv` that contains
# information about the subjects. The two files are linked through the
# `subjectID` column that should exist in both of them. For more information
# about this format, see also the `AFQ-Browser documentation
# <https://yeatmanlab.github.io/AFQ-Browser/dataformat.html>`_ (items 2 and 3).

workdir = download_weston_havens()

##########################################################################
# Read in the data
# ----------------
# Next, we read in the data. The :func:`load_afq_data` function expects a string
# input that points to a directory that holds appropriately-shaped data and
# returns variables that we will use below in our analysis of the data.

afqdata = load_afq_data(
    fn_nodes=op.join(workdir, "nodes.csv"),
    fn_subjects=op.join(workdir, "subjects.csv"),
    dwi_metrics=["md", "fa"],
    target_cols=["Age"],
)

# afqdata is a namedtuple. You can access it's fields using dot notation or by
# unpacking the tuple. To see all of the available fields use `afqdata._fields`
X = afqdata.X
y = afqdata.y
groups = afqdata.groups
feature_names = afqdata.feature_names
group_names = afqdata.group_names
subjects = afqdata.subjects

##########################################################################
# Create an analysis pipeline
# ---------------------------
#

pipe = make_afq_regressor_pipeline(
    imputer_kwargs={"strategy": "median"},  # Use median imputation
    use_cv_estimator=True,  # Automatically determine the best hyperparameters
    scaler="standard",  # Standard scale the features before regression
    groups=groups,
    verbose=0,  # Be quiet!
    pipeline_verbosity=False,  # No really, be quiet!
    tuning_strategy="bayes",  # Use BayesSearchCV to determine the optimal hyperparameters
    n_bayes_iter=10,  # Consider only this many points in hyperparameter space
    cv=3,  # Use three CV splits to evaluate each hyperparameter combination
    l1_ratio=[0.0, 1.0],  # Explore the entire range of ``l1_ratio``
    eps=5e-2,  # This is the ratio of the smallest to largest ``alpha`` value
    tol=1e-2,  # Set a lenient convergence tolerance just for this example
)

# ``pipe`` is a scikit-learn pipeline and can be used in other scikit-learn
# functions. For example, here we are doing 5-fold cross-validation using scikit
# learn's :func:`cross_validate` function.

scores = cross_validate(
    pipe, X, y, cv=5, return_train_score=True, return_estimator=True
)

print(f"Mean train score: {np.mean(scores['train_score']):5.3f}")
print(f"Mean test score:  {np.mean(scores['test_score']):5.3f}")
print(f"Mean fit time:    {np.mean(scores['fit_time']):5.2f}s")
print(f"Mean score time:  {np.mean(scores['score_time']):5.2f}s")

mean_coefs = np.mean(
    np.abs([est.named_steps["estimate"].coef_ for est in scores["estimator"]]), axis=0
)

fig, ax = plt.subplots(1, 1, figsize=(8, 8))
_ = ax.plot(mean_coefs[1800:], color="black", lw=2)
_ = ax.set_xlim(0, 1800)

colors = plt.get_cmap("tab20").colors
for grp, grp_name, color in zip(groups[:18], group_names[18:], colors):
    _ = ax.axvspan(grp.min(), grp.max() + 1, color=color, alpha=0.8, label=grp_name[1])

box = ax.get_position()
ax.set_position([box.x0, box.y0 + box.height * 0.375, box.width, box.height * 0.625])

_ = ax.legend(loc="upper center", bbox_to_anchor=(0.5, -0.125), ncol=3)
_ = ax.set_ylabel(r"$\hat{\beta}$", fontsize=16)
_ = ax.set_xlabel("Group principal component", fontsize=16)
_ = ax.set_title("Group Principal Regression Coefficients (MD only)", fontsize=18)



"""
=====================================
Load and interact with an AFQ dataset
=====================================

This example loads AFQ data from CSV files and manipulates that data using
scikit-learn transformers and estimators. First we fetch the Weston-Havens
dataset described in Yeatman et al [1]_. This dataset contains tractometry
features from 77 subjects ages 6-50.

Next, we split the dataset into a train and test split, impute missing values,
and fit a LASSO model, all using :class:`AFQDataset` methods. Predictive
performance for the default LASSO model is abysmal. It is only used here to
demonstrate the use of scikit-learn estimators. In a research setting, one might
use more advanced estimators, such as the SGL [2]_, a gradient boosting machine,
or a neural network.

Finally, we convert the AFQDataset to a tensorflow dataset and fit a basic
one-dimensional CNN to predict age from the features. This last step requires that
AFQ-Insight has been installed with::

    pip install afqinsight[tf]

or that tensorflow has been separately installed with::

    pip install tensorflow

.. [1]  Jason D. Yeatman, Brian A. Wandell, & Aviv A. Mezer, "Lifespan
    maturation and degeneration of human brain white matter" Nature
    Communications, vol. 5:1, pp. 4932, 2014 DOI: 10.1038/ncomms5932

.. [2]  Adam Richie-Halford, Jason Yeatman, Noah Simon, and Ariel Rokem
   "Multidimensional analysis and detection of informative features in human
   brain white matter" PLOS Computational Biology, 2021 DOI:
   10.1371/journal.pcbi.1009136

"""
import afqinsight.nn.tf_models as nn
import os.path as op
import tensorflow as tf

from afqinsight.datasets import download_weston_havens
from afqinsight import AFQDataset

from sklearn.impute import SimpleImputer
from sklearn.linear_model import Lasso
from sklearn.metrics import r2_score
from sklearn.model_selection import train_test_split

##########################################################################
# Fetch example data
# ------------------
#
# The :func:`download_weston_havens` function download the data used in this
# example and places it in the `~/.cache/afq-insight/weston_havens` directory.
# If the directory does not exist, it is created. The data follows the format
# expected by the :func:`load_afq_data` function: a file called `nodes.csv` that
# contains AFQ tract profiles and a file called `subjects.csv` that contains
# information about the subjects. The two files are linked through the
# `subjectID` column that should exist in both of them. For more information
# about this format, see also the `AFQ-Browser documentation
# <https://yeatmanlab.github.io/AFQ-Browser/dataformat.html>`_ (items 2 and 3).

workdir = download_weston_havens()

##########################################################################
# Read in the data
# ----------------
#
# Next, we read in the data. The :func:`AFQDataset.from_files` static method
# expects a the filenames of a nodes.csv and subjects.csv file, and returns a
# dataset object.

dataset = AFQDataset.from_files(
    fn_nodes=op.join(workdir, "nodes.csv"),
    fn_subjects=op.join(workdir, "subjects.csv"),
    dwi_metrics=["md", "fa"],
    target_cols=["Age"],
)

##########################################################################
# Train / test split
# ------------------
#
# We can use the dataset in the :func:`train_test_split` function just as we
# would with an array.

dataset_train, dataset_test = train_test_split(dataset, test_size=1 / 3)

##########################################################################
# Impute missing values
# ---------------------
#
# Next we train an imputer on the training set and use it to transform the
# features in both the training and the test set.

imputer = dataset_train.model_fit(SimpleImputer(strategy="median"))
dataset_train = dataset_train.model_transform(imputer)
dataset_test = dataset_test.model_transform(imputer)

##########################################################################
# Fit a LASSO model
# -----------------
#
# Next we fit a LASSO estimator to the training data and print the score of that
# model on the test dataset.

estimator = dataset_train.model_fit(Lasso())
y_pred = dataset_test.model_predict(estimator)
train_score = dataset_train.model_score(estimator)
test_score = dataset_test.model_score(estimator)
print("LASSO train score:", train_score)
print("LASSO test score: ", test_score)

##########################################################################
# Convert to tensorflow datasets
# ------------------------------
#
# Next we convert the train and test datasets to tensorflow datasets
# and use one of AFQ-Insight's built-in one-dimensional CNNs to predict
# age. This part of the example will only work if you have either installed
# AFQ-Insight with tensorflow using::
#
#     pip install afqinsight[tf]
#
# or separately install tensorflow using::
#
#     pip install tensorflow
#
# This model also performs poorly. It turns out predicting age in
# this dataset requires a bit more work.

tfset_train = dataset_train.as_tensorflow_dataset()
tfset_test = dataset_test.as_tensorflow_dataset()

batch_size = 2
tfset_train = tfset_train.batch(8)
tfset_test = tfset_test.batch(8)

print("CNN Architecture")
model = nn.cnn_lenet(
    input_shape=(100, 40), output_activation=None, n_classes=1, verbose=True
)

model.compile(
    loss="mean_squared_error",
    optimizer=tf.keras.optimizers.Adam(learning_rate=1e-4),
    metrics=["mean_squared_error"],
)

model.fit(tfset_train, epochs=500, validation_data=tfset_test, verbose=0)

print()
print("CNN R^2 score: ", r2_score(dataset_test.y, model.predict(tfset_test)))



import afqinsight.augmentation as aug
import numpy as np
import os.path as op
import pytest

from afqinsight.datasets import bundles2channels, download_sarica, AFQDataset
from inspect import signature
from sklearn.impute import SimpleImputer


@pytest.fixture
def sarica_Xy():
    sarica_dir = download_sarica()
    afq_data = AFQDataset.from_files(
        fn_nodes=op.join(sarica_dir, "nodes.csv"),
        fn_subjects=op.join(sarica_dir, "subjects.csv"),
        dwi_metrics=["md", "fa"],
        target_cols=["class"],
        label_encode_cols=["class"],
    )
    X = SimpleImputer(strategy="median").fit_transform(afq_data.X)
    X = bundles2channels(X, n_nodes=100, n_channels=40, channels_last=True)
    return X, afq_data.y


RANDOM_AUGMENTATION_METHODS = [
    aug.jitter,
    aug.scaling,
    aug.permutation,
    aug.magnitude_warp,
    aug.time_warp,
    aug.window_slice,
    aug.window_warp,
]

PATTERN_AUGMENTATION_METHODS = [
    aug.spawner,
    aug.wdba,
    aug.random_guided_warp,
    aug.discriminative_guided_warp,
]


@pytest.mark.parametrize("aug_method", RANDOM_AUGMENTATION_METHODS)
def test_random_smoke(sarica_Xy, aug_method):
    X, _ = sarica_Xy
    X_aug = aug_method(X)
    assert X.shape == X_aug.shape  # nosec

    parameters = signature(aug_method).parameters
    if "warp_channels_independently" in parameters:
        X_aug = aug_method(X, warp_channels_independently=True)
        assert X.shape == X_aug.shape  # nosec

    if "seg_mode" in parameters:
        X_aug = aug_method(X, seg_mode="random")
        assert X.shape == X_aug.shape  # nosec


@pytest.mark.parametrize("aug_method", PATTERN_AUGMENTATION_METHODS)
def test_pattern_smoke(sarica_Xy, aug_method):
    X, y = sarica_Xy
    X_aug = aug_method(x=X, labels=y)
    assert X.shape == X_aug.shape  # nosec

    y[10] = 2.0
    X_aug = aug_method(x=X, labels=y)
    assert X.shape == X_aug.shape  # nosec

    parameters = signature(aug_method).parameters
    if "dtw_type" in parameters:
        X_aug = aug_method(x=X, labels=y, dtw_type="shape")
        assert X.shape == X_aug.shape  # nosec


MEAN_ZERO_METHODS = [aug.jitter, aug.scaling, aug.magnitude_warp, aug.time_warp]


@pytest.mark.parametrize("aug_method", MEAN_ZERO_METHODS)
def test_augmentation_average(sarica_Xy, aug_method):
    n_aug = 10
    X, _ = sarica_Xy
    sigma = 1e-2
    aug_mean = np.mean(
        np.stack([aug_method(X, sigma=sigma) for _ in range(n_aug)]), axis=0
    )
    assert np.linalg.norm(X - aug_mean) / np.linalg.norm(X) <= sigma



"""Perform statistical matching comparing test and control units."""

import numpy as np
import pandas as pd
from scipy.optimize import linear_sum_assignment


__all__ = ["mahalonobis_dist_match"]


def mahalonobis_dist_match(
    data=None, test=None, ctrl=None, status_col=None, feature_cols=None, threshold=0.2
):
    """
    Perform Mahalanobis distance matching (MDM) on a pandas dataframe or 2d array.

    If data is None, then test and ctrl must be arrays.
    If data is provided, then status_col and feature_cols are used.

    Parameters
    ----------
    data : pandas Dataframe, optional
        The feature samples for test subjects.
        Default: None
    test : array-like of shape (n_samples1, n_features), optional
        The feature samples for test subjects.
        Default: None
    ctrl : array-like of shape (n_samples2, n_features), optional
        Feature samples for control subjects from which
        matches will be selected.
        Default: None
    status_col : str, optional
        Column with subject status. Must have only two unique values,
        evaluating to True and False, where True indicates a test subject
        and False indicates a control.
        Default: None
    feature_cols : list of str, optional
        List of columns to use as features for MDM.
        If feature_columns is None then we use all of the columns
        except status_col as the features.
        Default: None
    threshold : float, optional
        Mahalnobis distance beyond which matches are not accepted.
        Default: None

    Returns
    -------
    selected_data : pandas Dataframe
        If data is None, returns a Dataframe with columns named
        'feature_0', 'feature_1', etc. and a 'status' column
        where 1 corresponds to test and 0 corresponds to control.
        If data is not None, returns data with only the selected rows.
    """
    if data is not None:
        if status_col is None:
            raise ValueError(
                ("If data is provided," " status_col must also be provided")
            )

        if len(data[status_col].unique()) != 2:
            raise ValueError(
                ("status_col in data must " "contain only 2 unique values")
            )

        if feature_cols is None:
            feature_cols = data.columns.drop(status_col)

        df_test_rows = data[status_col].astype(bool)
        df_ctrl_rows = ~data[status_col].astype(bool)
        test = data.loc[df_test_rows, feature_cols].to_numpy()
        ctrl = data.loc[df_ctrl_rows, feature_cols].to_numpy()
    else:
        test = np.asarray(test)
        ctrl = np.asarray(ctrl)

    if np.any(np.isnan(test)) or np.any(np.isnan(ctrl)):
        raise ValueError(
            (
                "There are NaNs in test or ctrl data. "
                "Please replace these NaNs using interpolation or by removing "
                "the subjects with NaNs before calling mahalonobis_dist_match. "
            )
        )

    # calculate Mahalonobis distance between test and control
    nbrs = _mahalonobis_dist(test, ctrl)

    # assign neighbors using Munkres algorithm
    row_ind, col_ind = linear_sum_assignment(nbrs)

    # remove matches that are too bad
    filtered_test_ind = []
    filtered_ctrl_ind = []
    for row_idx, col_idx in zip(row_ind, col_ind):
        if nbrs[row_idx, col_idx] <= threshold:
            filtered_test_ind.append(row_idx)
            filtered_ctrl_ind.append(col_idx)

    # contstruct dataframe if no dataframe is provided
    if data is None:
        data = {}
        filtered_test = test[filtered_test_ind]
        filtered_ctrl = ctrl[filtered_ctrl_ind]
        for i in range(test.shape[1]):
            data[f"feature_{i}"] = np.concatenate(
                (filtered_test[:, i], filtered_ctrl[:, i]), axis=0
            )
        data["status"] = np.concatenate(
            (np.ones(len(filtered_test_ind)), np.zeros(len(filtered_ctrl_ind))), axis=0
        )
        return pd.DataFrame(data)
    else:
        # determine selected rows
        df_test_idx = np.argwhere(df_test_rows.to_numpy()).flatten()
        df_ctrl_idx = np.argwhere(df_ctrl_rows.to_numpy()).flatten()
        all_idx = np.concatenate(
            (df_test_idx[filtered_test_ind], df_ctrl_idx[filtered_ctrl_ind]), axis=0
        )
        return data.iloc[all_idx]


def _mahalonobis_dist(arr1, arr2):
    """
    Calculate the Mahalonobis distance between two 2d arrays along the first axis.

    Parameters
    ----------
    arr1 : array-like of shape (n_samples1, n_features)
        First input array.
    arr2 : array-like of shape (n_samples2, n_features)
        Second input array.

    Returns
    -------
    nbrs : array-like of shape (n_samples1, n_samples2)
        Mahalonobis distance between each sample in the
        two input arrays.
    """
    v_inv = np.linalg.inv(np.cov(np.concatenate((arr1, arr2), axis=0).T, ddof=0))
    diff = arr1[:, np.newaxis] - arr2[np.newaxis, :]
    nbrs = np.sqrt(np.sum((diff @ v_inv) * diff, axis=2))
    return nbrs
